""" Create a samplesheet for Novaseq flowcells """

import logging
from typing import Dict, List, Set

from cg.apps.demultiplex.sample_sheet import index
from cg.apps.demultiplex.sample_sheet.dummy_sample import dummy_sample
from cg.apps.demultiplex.sample_sheet.index import Index
from cg.apps.lims.samplesheet import LimsFlowcellSample
from cg.models.demultiplex.run_parameters import RunParameters
from cgmodels.demultiplex.sample_sheet import get_sample_sheet

LOG = logging.getLogger(__name__)


class SampleSheetCreator:
    """ Create a raw sample sheet for Novaseq flowcells """

    HEADER_TO_LIMS_KEY = {
        "FCID": "flowcell_id",
        "Lane": "lane",
        "SampleID": "sample_id",
        "SampleRef": "sample_ref",
        "index": "index",
        "index2": "index2",
        "SampleName": "sample_name",
        "Control": "control",
        "Recipe": "recipe",
        "Operator": "operator",
        "Project": "project",
    }
    SAMPLE_SHEET_HEADERS = [
        "FCID",
        "Lane",
        "SampleID",
        "SampleRef",
        "index",
        "index2",
        "SampleName",
        "Control",
        "Recipe",
        "Operator",
        "Project",
    ]

    def __init__(
        self,
        flowcell_id: str,
        lims_samples: List[LimsFlowcellSample],
        run_parameters: RunParameters,
    ):
        self.flowcell_id: str = flowcell_id
        self.lims_samples: List[LimsFlowcellSample] = lims_samples
        self.run_parameters: RunParameters = run_parameters

    @property
    def valid_indexes(self) -> List[Index]:
        return index.get_valid_indexes(dual_indexes_only=True)

    @staticmethod
    def get_project_name(project: str) -> str:
        """ Only keeps the first part of the project name """
        return project.split()[0]

    def add_dummy_samples(self) -> None:
        """Add all dummy samples with non existing indexes to samples

        dummy samples are added if there are indexes that are not used by the actual samples.
        This means that we will add each dummy sample (that is needed) to each lane
        """
        LOG.info("Adding dummy samples for unused indexes")
        indexes_by_lane: Dict[int, Set[str]] = index.get_indexes_by_lane(samples=self.lims_samples)
        for lane, lane_indexes in indexes_by_lane.items():
            LOG.debug("Add dummy samples for lane %s", lane)
            for index_obj in self.valid_indexes:
                if index.index_exists(index=index_obj.sequence, indexes=lane_indexes):
                    LOG.debug("Index %s already in use", index_obj.sequence)
                    continue
                dummy_sample_obj: LimsFlowcellSample = dummy_sample(
                    flowcell=self.flowcell_id,
                    dummy_index=index_obj.sequence,
                    lane=lane,
                    name=index_obj.name,
                )
                LOG.debug("Adding dummy sample %s to lane %s", dummy_sample_obj, lane)
                self.lims_samples.append(dummy_sample_obj)

    def remove_unwanted_samples(self) -> None:
        """ Filter out samples with indexes of unwanted length and single indexes """
        LOG.info("Removing all samples without dual indexes")
        samples_to_keep = []
        sample: LimsFlowcellSample
        for sample in self.lims_samples:
            if not index.is_dual_index(sample.index):
                LOG.warning("Removing sample %s since it does not have dual index", sample)
                continue
            samples_to_keep.append(sample)
        self.lims_samples = samples_to_keep

    @staticmethod
    def convert_sample_to_header_dict(
        sample: LimsFlowcellSample, sample_sheet_headers: List[str], header_to_lims: Dict[str, str]
    ) -> List[str]:
        """Convert a lims sample object to a dict with keys that corresponds to the sample sheet headers"""
        LOG.debug("Use sample sheet header %s", sample_sheet_headers)
        sample_dict = sample.dict()
        return [str(sample_dict[header_to_lims[header]]) for header in sample_sheet_headers]

    def convert_to_sample_sheet(self) -> str:
        """Convert all samples to a string with the sample sheet"""
        LOG.info("Convert samples to string")
        sample_sheet = ["[Data]", ",".join(self.SAMPLE_SHEET_HEADERS)]
        for sample in self.lims_samples:
            sample_sheet.append(
                ",".join(
                    self.convert_sample_to_header_dict(
                        sample=sample,
                        sample_sheet_headers=self.SAMPLE_SHEET_HEADERS,
                        header_to_lims=self.HEADER_TO_LIMS_KEY,
                    )
                )
            )
        return "\n".join(sample_sheet)

    def construct_sample_sheet(self) -> str:
        """ Construct the sample sheet """
        LOG.info("Constructing sample sheet for %s", self.flowcell_id)
        # Create dummy samples for the indexes that is missing
        if self.run_parameters.run_type == "wgs":
            self.add_dummy_samples()
        else:
            LOG.info("Skip adding dummy samples since run is not WGS")
        self.remove_unwanted_samples()
        index.adapt_indexes(
            samples=self.lims_samples,
            control_software_version=self.run_parameters.control_software_version,
            reagent_kit_version=self.run_parameters.reagent_kit_version,
            expected_index_length=self.run_parameters.index_length,
        )
        sample_sheet: str = self.convert_to_sample_sheet()
        LOG.info("Validating sample sheet")
        get_sample_sheet(sample_sheet=sample_sheet, sheet_type="S2")
        LOG.info("Sample sheet looks fine")
        return sample_sheet



"""Fixtures for deliver commands"""

from pathlib import Path

import pytest
from tests.store_helpers import StoreHelpers

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.store import Store
from housekeeper.store import models as hk_models

# Paths


@pytest.fixture(name="delivery_inbox")
def fixture_delivery_inbox(project_dir: Path, customer_id: Path, ticket_nr: int) -> Path:
    return project_dir / customer_id / "inbox" / str(ticket_nr)


@pytest.fixture(name="deliver_vcf_path")
def fixture_deliver_vcf_path(
    delivery_inbox: Path, family_name: str, case_id: str, vcf_file: Path
) -> Path:
    return delivery_inbox / family_name / vcf_file.name.replace(case_id, family_name)


@pytest.fixture(name="base_context")
def fixture_base_context(
    base_store: Store, real_housekeeper_api: HousekeeperAPI, project_dir: Path
) -> dict:
    return {
        "status_db": base_store,
        "housekeeper_api": real_housekeeper_api,
        "delivery_path": str(project_dir),
    }


@pytest.fixture(name="mip_delivery_bundle")
def fixture_mip_delivery_bundle(
    case_hk_bundle_no_files: dict, sample1_cram: Path, vcf_file: Path
) -> dict:
    """Return a bundle that includes files used when delivering MIP analysis data"""
    case_hk_bundle_no_files["files"] = [
        {"path": str(sample1_cram), "archive": False, "tags": ["ADM1", "cram"]},
        {"path": str(vcf_file), "archive": False, "tags": ["vcf-snv-clinical"]},
    ]
    return case_hk_bundle_no_files


@pytest.fixture(name="mip_dna_housekeeper")
def fixture_mip_dna_housekeeper(
    real_housekeeper_api: HousekeeperAPI, mip_delivery_bundle: dict, helpers: StoreHelpers
) -> HousekeeperAPI:
    helpers.ensure_hk_bundle(real_housekeeper_api, bundle_data=mip_delivery_bundle)
    # assert that the files exists
    version_obj: hk_models.Version = real_housekeeper_api.last_version(mip_delivery_bundle["name"])
    real_housekeeper_api.include(version_obj=version_obj)

    return real_housekeeper_api


@pytest.fixture(name="populated_mip_context")
def fixture_populated_mip_context(
    analysis_store: Store, mip_dna_housekeeper: HousekeeperAPI, project_dir: Path
) -> dict:
    return {
        "status_db": analysis_store,
        "housekeeper_api": mip_dna_housekeeper,
        "delivery_path": str(project_dir),
    }



""" Test adding files with housekeeper api """


def test_add_file_with_flat_tag(housekeeper_api, helpers, hk_bundle_data, microbial_orderform):
    """Test that we can call hk with one existing tag"""

    # GIVEN an hk api populated with a version obj
    version_obj = helpers.ensure_hk_version(housekeeper_api, hk_bundle_data)
    # GIVEN a tag that is just a string that exists
    tag_name = "bed"
    assert housekeeper_api.tag(tag_name)

    # WHEN we call add_file
    new_file = housekeeper_api.add_file(microbial_orderform, version_obj, tag_name)

    # THEN the file should have been added to hk
    assert new_file


def test_add_file_with_list_of_tags(housekeeper_api, helpers, hk_bundle_data, microbial_orderform):
    """Test that we can call hk with more than one tags"""

    # GIVEN an hk api populated with a version obj
    version_obj = helpers.ensure_hk_version(housekeeper_api, hk_bundle_data)
    # GIVEN a list of tags that does not exist
    tags = ["missing_tag_1", "missing_tag_2"]
    for tag_name in tags:
        assert housekeeper_api.tag(tag_name) is None

    # WHEN we call add_file
    new_file = housekeeper_api.add_file(microbial_orderform, version_obj, tags)

    # THEN the file should have been added to hk
    assert new_file



"""Code for uploading to scout via CLI"""
import logging
import sys
from pathlib import Path
from typing import Optional

import click
from cg.meta.upload.scout.scoutapi import UploadScoutAPI
from cg.models.scout.scout_load_config import ScoutLoadConfig
from cg.store.models import Family
from housekeeper.store import models as hk_models
from ruamel.yaml import YAML

from ...meta.workflow.mip_dna import MipDNAAnalysisAPI
from .utils import suggest_cases_to_upload

LOG = logging.getLogger(__name__)


@click.command()
@click.option("-r", "--re-upload", is_flag=True, help="re-upload existing analysis")
@click.option("-p", "--print", "print_console", is_flag=True, help="print config values")
@click.argument("case_id", required=False)
@click.pass_context
def scout(context, re_upload: bool, print_console: bool, case_id: str):
    """Upload variants from analysis to Scout."""

    if not context.obj.get("analysis_api"):
        context.obj["analysis_api"] = MipDNAAnalysisAPI(context.obj)

    LOG.info("----------------- SCOUT -----------------------")

    if not case_id:
        suggest_cases_to_upload(context)
        return

    context.invoke(
        create_scout_load_config, case_id=case_id, print_console=print_console, re_upload=re_upload
    )
    if not print_console:
        context.invoke(upload_case_to_scout, case_id=case_id, re_upload=re_upload)


@click.command(name="create-scout-load-config")
@click.argument("case-id")
@click.option("-p", "--print", "print_console", is_flag=True, help="Only print config")
@click.option("-r", "--re-upload", is_flag=True, help="Delete existing load configs")
@click.pass_context
def create_scout_load_config(context, case_id: str, print_console: bool, re_upload: bool):
    """Create a load config for a case in scout and add it to housekeeper"""

    if not context.obj.get("analysis_api"):
        context.obj["analysis_api"] = MipDNAAnalysisAPI(context.obj)
    analysis_api = context.obj["analysis_api"]
    scout_upload_api: UploadScoutAPI = context.obj["scout_upload_api"]

    LOG.info("----------------- CREATE CONFIG -----------------------")

    LOG.info("Fetching family object")
    case_obj: Family = analysis_api.status_db.family(case_id)
    LOG.info("Create load config")
    if not case_obj.analyses:
        LOG.warning("Could not find analyses for %s", case_id)
        raise click.Abort
    try:
        scout_load_config: ScoutLoadConfig = scout_upload_api.generate_config(case_obj.analyses[0])
    except SyntaxError as err:
        LOG.warning("%s", err)
        raise click.Abort
    LOG.info("Found load config %s", scout_load_config)
    analysis_context: str = "mip-rd-dna"
    if scout_load_config.track == "cancer":
        analysis_context = "balsamic"
    root_dir: Path = Path(context.obj[analysis_context]["root"])
    LOG.info("Set root dir to %s", root_dir)
    file_path: Path = root_dir / case_id / "scout_load.yaml"

    if print_console:
        yaml = YAML()
        yaml.indent(mapping=4, sequence=6, offset=3)
        yaml.dump(scout_load_config.dict(exclude_none=True), sys.stdout)
        LOG.info("Would save file to %s", file_path)
        return

    if file_path.exists():
        LOG.warning("Scout load config %s already exists", file_path)
        if re_upload:
            LOG.info("Deleting old load config")
            file_path.unlink()
        else:
            LOG.info(
                "You might remove the file and try again, consider that you might also have it in housekeeper"
            )
            raise click.Abort
    LOG.info("Saving config file to disc")
    scout_upload_api.save_config_file(upload_config=scout_load_config, file_path=file_path)

    try:
        LOG.info("Add config file to housekeeper")
        scout_upload_api.add_scout_config_to_hk(
            config_file_path=file_path, case_id=case_id, delete=re_upload
        )
    except FileExistsError as err:
        LOG.warning("%s, consider removing the file from housekeeper and try again", str(err))
        raise click.Abort


@click.command(name="upload-case-to-scout")
@click.option("-r", "--re-upload", is_flag=True, help="re-upload existing analysis")
@click.option("--dry-run", is_flag=True)
@click.argument("case_id")
@click.pass_context
def upload_case_to_scout(context, re_upload: bool, dry_run: bool, case_id: str):
    """Upload variants and case from analysis to Scout."""

    LOG.info("----------------- UPLOAD -----------------------")

    if not context.obj.get("analysis_api"):
        context.obj["analysis_api"] = MipDNAAnalysisAPI(context.obj)
    analysis_api = context.obj["analysis_api"]

    tag_name = UploadScoutAPI.get_load_config_tag()
    version_obj = analysis_api.housekeeper_api.last_version(case_id)
    scout_config_file: Optional[
        hk_models.File
    ] = analysis_api.housekeeper_api.fetch_file_from_version(
        version_obj=version_obj, tags={tag_name}
    )

    if scout_config_file is None:
        raise FileNotFoundError(f"No scout load config was found in housekeeper for {case_id}")

    LOG.info("uploading case %s to scout", case_id)

    if not dry_run:
        analysis_api.scout_api.upload(
            scout_load_config=scout_config_file.full_path, force=re_upload
        )

    LOG.info("uploaded to scout using load config %s", scout_config_file.full_path)
    LOG.info("Case loaded succesfully to Scout")



from cg.apps.mip import parse_qcmetrics, parse_sampleinfo


def parse_mip_analysis(mip_config_raw: dict, qcmetrics_raw: dict, sampleinfo_raw: dict) -> dict:
    """Parse the output analysis files from MIP for adding info
    to trend database

    Args:
        mip_config_raw (dict): raw YAML input from MIP analysis config file
        qcmetrics_raw (dict): raw YAML input from MIP analysis qc metric file
        sampleinfo_raw (dict): raw YAML input from MIP analysis qc sample info file
    Returns:
        dict: parsed data
    """
    outdata = {
        "analysis_sex": {},
        "at_dropout": {},
        "case": None,
        "duplicates": {},
        "gc_dropout": {},
        "genome_build": None,
        "rank_model_version": None,
        "insert_size_standard_deviation": {},
        "mapped_reads": {},
        "median_insert_size": {},
        "mip_version": None,
        "sample_ids": [],
    }

    _config(mip_config_raw=mip_config_raw, outdata=outdata)
    _qc_metrics(qcmetrics_raw=qcmetrics_raw, outdata=outdata)
    _qc_sample_info(sampleinfo_raw=sampleinfo_raw, outdata=outdata)

    return outdata


def _add_rank_model_version(outdata: dict, sampleinfo_data: dict) -> None:
    outdata["sv_rank_model_version"] = sampleinfo_data["sv_rank_model_version"]
    outdata["rank_model_version"] = sampleinfo_data["rank_model_version"]


def _qc_sample_info(outdata: dict, sampleinfo_raw: dict) -> None:
    sampleinfo_data = _parse_qc_sample_info_file(sampleinfo_raw)
    _add_mip_version(outdata, sampleinfo_data)
    _add_genome_build(outdata, sampleinfo_data)
    _add_rank_model_version(outdata, sampleinfo_data)


def _qc_metrics(outdata: dict, qcmetrics_raw: dict) -> None:
    qcmetrics_data = _parse_qc_metric_file_into_dict(qcmetrics_raw)
    _add_sample_level_info_from_qc_metric_file(outdata, qcmetrics_data)


def _config(mip_config_raw: dict, outdata: dict) -> dict:
    config_data = _parse_raw_mip_config_into_dict(mip_config_raw)
    _add_case_id(config_data, outdata)
    _add_all_samples_from_mip_config(config_data, outdata)
    return config_data


def _add_genome_build(outdata: dict, sampleinfo_data: dict) -> None:
    outdata["genome_build"] = sampleinfo_data["genome_build"]


def _add_mip_version(outdata: dict, sampleinfo_data: dict) -> None:
    outdata["mip_version"] = sampleinfo_data["version"]


def _parse_qc_sample_info_file(sampleinfo_raw: dict) -> dict:
    return parse_sampleinfo.parse_sampleinfo(sampleinfo_raw)


def _add_sample_level_info_from_qc_metric_file(outdata: dict, qcmetrics_data: dict) -> None:
    for sample_data in qcmetrics_data:
        _add_duplicate_reads(outdata, qcmetrics_data[sample_data])
        _add_mapped_reads(outdata, qcmetrics_data[sample_data])
        _add_predicted_sex(outdata, qcmetrics_data[sample_data])
        _add_dropout_rates(outdata, qcmetrics_data[sample_data])
        _add_insert_size_metrics(outdata, qcmetrics_data[sample_data])


def _add_dropout_rates(outdata: dict, sample_data: dict) -> None:
    outdata["at_dropout"][sample_data["id"]] = sample_data["at_dropout"]
    outdata["gc_dropout"][sample_data["id"]] = sample_data["gc_dropout"]


def _add_insert_size_metrics(outdata: dict, sample_data: dict) -> None:
    outdata["median_insert_size"][sample_data["id"]] = sample_data["median_insert_size"]
    outdata["insert_size_standard_deviation"][sample_data["id"]] = sample_data[
        "insert_size_standard_deviation"
    ]


def _add_predicted_sex(outdata: dict, sample_data: dict) -> None:
    outdata["analysis_sex"][sample_data["id"]] = sample_data["predicted_sex"]


def _add_mapped_reads(outdata: dict, sample_data: dict) -> None:
    mapped_reads_percent = sample_data["mapped"] * 100
    outdata["mapped_reads"][sample_data["id"]] = mapped_reads_percent


def _add_duplicate_reads(outdata: dict, sample_data: dict) -> None:
    duplicates_percent = sample_data["duplicates"] * 100
    outdata["duplicates"][sample_data["id"]] = duplicates_percent


def _parse_qc_metric_file_into_dict(qcmetrics_raw: dict) -> dict:
    qcmetrics_data = parse_qcmetrics.parse_qcmetrics(qcmetrics_raw)
    return qcmetrics_data


def _parse_raw_mip_config_into_dict(mip_config_raw: dict) -> dict:
    config_data = parse_sampleinfo.parse_config(mip_config_raw)
    return config_data


def _add_all_samples_from_mip_config(config_data: dict, outdata: dict) -> None:
    for sample_data in config_data["samples"]:
        outdata["sample_ids"].append(sample_data["id"])


def _add_case_id(config_data: dict, outdata: dict) -> None:
    outdata["case"] = config_data["case"]



import shutil
from pathlib import Path
from typing import Dict

import pytest
from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.apps.lims import LimsAPI
from cg.models.demultiplex.flowcell import Flowcell
from cg.utils import Process
from click.testing import CliRunner
from tests.apps.crunchy.conftest import fixture_sbatch_process
from tests.apps.demultiplex.conftest import (
    fixture_demultiplex_fixtures,
    fixture_lims_novaseq_samples,
    fixture_lims_novaseq_samples_file,
    fixture_novaseq_dir,
    fixture_novaseq_run_parameters,
    fixture_raw_samples_dir,
)


@pytest.fixture(name="novaseq_sample_sheet_path")
def fixture_novaseq_sample_sheet_path(demultiplex_fixtures: Path) -> Path:
    return demultiplex_fixtures / "SampleSheetS2.csv"


@pytest.fixture(name="flowcell_working_directory")
def fixture_flowcell_working_directory(novaseq_dir: Path, project_dir: Path) -> Path:
    """Return the path to a working directory that will be deleted after test is run

    This is a path to a flowcell directory with the run parameters present
    """
    working_dir: Path = project_dir / novaseq_dir.name
    working_dir.mkdir(parents=True)
    existing_flowcell: Flowcell = Flowcell(flowcell_path=novaseq_dir)
    working_flowcell: Flowcell = Flowcell(flowcell_path=working_dir)
    shutil.copy(
        str(existing_flowcell.run_parameters_path), str(working_flowcell.run_parameters_path)
    )
    return working_dir


@pytest.fixture(name="demultiplex_ready_flowcell")
def fixture_demultiplex_ready_flowcell(flowcell_working_directory: Path, novaseq_dir: Path) -> Path:
    """Return the path to a working directory that is ready for demultiplexing

    This is a path to a flowcell directory with all the files necessary to start demultiplexing present
    """
    existing_flowcell: Flowcell = Flowcell(flowcell_path=novaseq_dir)
    working_flowcell: Flowcell = Flowcell(flowcell_path=flowcell_working_directory)
    shutil.copy(str(existing_flowcell.sample_sheet_path), str(working_flowcell.sample_sheet_path))
    working_flowcell.copy_complete_path.touch()
    working_flowcell.rta_complete_path.touch()
    return flowcell_working_directory


@pytest.fixture(name="sample_sheet_context")
def fixture_sample_sheet_context(lims_api: LimsAPI) -> dict:
    return {"lims_api": lims_api}


@pytest.fixture(name="demultiplex_configs")
def fixture_demultiplex_configs(project_dir: Path) -> dict:
    out_dir: Path = project_dir / "demultiplexed-runs"
    out_dir.mkdir(parents=True)
    return {
        "demultiplex": {
            "out_dir": str(out_dir),
            "slurm": {"account": "test", "mail_user": "testuser@github.se"},
        }
    }


@pytest.fixture(name="demultiplex_context")
def fixture_demultiplex_context(
    demultiplex_configs: dict, sbatch_process: Process
) -> Dict[str, DemultiplexingAPI]:
    demux_api = DemultiplexingAPI(config=demultiplex_configs)
    demux_api.slurm_api.process = sbatch_process
    return {"demultiplex_api": demux_api}


@pytest.fixture(name="cli_runner")
def fixture_cli_runner():
    """Create a CliRunner"""
    return CliRunner()



""" Start of CLI """
import logging
import sys

import cg
import click
import coloredlogs
import ruamel.yaml
from cg.cli.delete.base import delete
from cg.cli.set.base import set_cmd
from cg.cli.store.store import store as store_cmd
from cg.store import Store

from .add import add as add_cmd
from .backup import backup
from .clean import clean
from .compress.base import compress, decompress
from .deliver.base import deliver as deliver_cmd
from .demultiplex.base import demultiplex as demultiplex_cmd
from .deploy.base import deploy as deploy_cmd
from .export import export
from .get import get
from .import_cmd import import_cmd
from .reset import reset_cmd
from .status import status
from .transfer import transfer
from .upload.base import upload
from .workflow.base import workflow as workflow_cmd

LOG = logging.getLogger(__name__)
LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR"]


@click.group()
@click.option("-c", "--config", type=click.File(), help="path to config file")
@click.option("-d", "--database", help="path/URI of the SQL database")
@click.option(
    "-l", "--log-level", type=click.Choice(LEVELS), default="INFO", help="lowest level to log at"
)
@click.option("--verbose", is_flag=True, help="Show full log information, time stamp etc")
@click.version_option(cg.__version__, prog_name=cg.__title__)
@click.pass_context
def base(context, config, database, log_level, verbose):
    """cg - interface between tools at Clinical Genomics."""
    if verbose:
        log_format = "%(asctime)s %(hostname)s %(name)s[%(process)d] %(levelname)s %(message)s"
    else:
        log_format = "%(message)s" if sys.stdout.isatty() else None

    coloredlogs.install(level=log_level, fmt=log_format)
    context.obj = ruamel.yaml.safe_load(config) if config else {}
    if database:
        context.obj["database"] = database


@base.command()
@click.option("--reset", is_flag=True, help="reset database before setting up tables")
@click.option("--force", is_flag=True, help="bypass manual confirmations")
@click.pass_context
def init(context, reset, force):
    """Setup the database."""
    status_db = Store(context.obj["database"])
    existing_tables = status_db.engine.table_names()
    if force or reset:
        if existing_tables and not force:
            message = f"Delete existing tables? [{', '.join(existing_tables)}]"
            click.confirm(click.style(message, fg="yellow"), abort=True)
        status_db.drop_all()
    elif existing_tables:
        LOG.error("Database already exists, use '--reset'")
        context.abort()

    status_db.create_all()
    LOG.info("Success! New tables: %s", ", ".join(status_db.engine.table_names()))


base.add_command(add_cmd)
base.add_command(backup)
base.add_command(clean)
base.add_command(compress)
base.add_command(decompress)
base.add_command(delete)
base.add_command(export)
base.add_command(get)
base.add_command(import_cmd)
base.add_command(reset_cmd)
base.add_command(set_cmd)
base.add_command(status)
base.add_command(transfer)
base.add_command(upload)
base.add_command(workflow_cmd)
base.add_command(store_cmd)
base.add_command(deploy_cmd)
base.add_command(deliver_cmd)
base.add_command(demultiplex_cmd)



import logging
from pathlib import Path
from typing import List

from cg.apps.lims.samplesheet import LimsFlowcellSample
from cg.cli.demultiplex.sample_sheet import create_sheet
from cg.models.demultiplex.flowcell import Flowcell
from click import testing


def test_create_sample_sheet_no_run_parameters(
    cli_runner: testing.CliRunner,
    demultiplex_fixtures: Path,
    sample_sheet_context: dict,
    caplog,
    mocker,
):
    # GIVEN a folder with a non existing sample sheet
    flowcell_object: Flowcell = Flowcell(demultiplex_fixtures)
    assert flowcell_object.run_parameters_path.exists() is False
    mocker.patch("cg.cli.demultiplex.sample_sheet.flowcell_samples", return_value=[{"sample": 1}])

    # WHEN running the create sample sheet command
    result: testing.Result = cli_runner.invoke(
        create_sheet, [str(demultiplex_fixtures)], obj=sample_sheet_context
    )

    # THEN assert it exits with a non zero exit code
    assert result.exit_code != 0
    # THEN assert the correct information is communicated
    assert "Could not find run parameters file" in caplog.text


def test_create_sample_sheet(
    cli_runner: testing.CliRunner,
    flowcell_working_directory: Path,
    sample_sheet_context: dict,
    lims_novaseq_samples: List[LimsFlowcellSample],
    mocker,
):
    # GIVEN a flowcell directory with some run parameters
    flowcell: Flowcell = Flowcell(flowcell_working_directory)
    assert flowcell.run_parameters_path.exists()
    # GIVEN that there is no sample sheet present
    assert not flowcell.sample_sheet_exists()
    mocker.patch(
        "cg.cli.demultiplex.sample_sheet.flowcell_samples", return_value=lims_novaseq_samples
    )
    # GIVEN a lims api that returns some samples

    # WHEN creating a sample sheet
    result = cli_runner.invoke(
        create_sheet, [str(flowcell_working_directory)], obj=sample_sheet_context
    )

    # THEN assert it exits with success
    assert result.exit_code == 0
    # THEN assert that the sample sheet was created
    assert flowcell.sample_sheet_exists()
    # THEN assert that the sample sheet is on the correct format
    assert flowcell.validate_sample_sheet()



"""Priority specific constants"""

PRIORITY_MAP = {"research": 0, "standard": 1, "priority": 2, "express": 3, "clinical trials": 4}

REV_PRIORITY_MAP = {value: key for key, value in PRIORITY_MAP.items()}

PRIORITY_OPTIONS = list(PRIORITY_MAP.keys())



"""Constants specific for compression"""
import datetime

# Constants for crunchy
FASTQ_FIRST_READ_SUFFIX = "_R1_001.fastq.gz"
FASTQ_SECOND_READ_SUFFIX = "_R2_001.fastq.gz"
SPRING_SUFFIX = ".spring"

# Number of days until fastqs counts as old
FASTQ_DELTA = 21
# Get the fastq delta in datetime format
FASTQ_DATETIME_DELTA = datetime.timedelta(days=FASTQ_DELTA)

PROBLEMATIC_CASES = [
    "causalmite",
    "deepcub",
    "expertalien",
    "fluenteagle",
    "grandkoi",
    "lovingmayfly",
    "loyalegret",
    "modernbee",
    "proudcollie",
    "richalien",
    "suremako",
    "wisestork",
]

# List of cases used for validation that we should skip
VALIDATION_CASES = [
    "bigdrum",
    "bosssponge",
    "busycolt",
    "casualgannet",
    "civilkoala",
    "cleanshrimp",
    "daringpony",
    "easybeetle",
    "epicasp",
    "finequagga",
    "firstfawn",
    "fleetjay",
    "frankhusky",
    "gamedeer",
    "gladthrush",
    "helpedfilly",
    "hotskink",
    "hotviper",
    "intentcorgi",
    "intentmayfly",
    "keencalf",
    "keenviper",
    "lightprawn",
    "livingox",
    "luckyhog",
    "meetpossum",
    "newaphid",
    "mintbaboon",
    "mintyeti",
    "moralcattle",
    "moralgoat",
    "onemite",
    "proeagle",
    "propercoral",
    "pumpedcat",
    "rightmacaw",
    "safeguinea",
    "sharpparrot",
    "sharppigeon",
    "strongbison",
    "strongman",
    "topsrhino",
    "unitedbeagle",
    "usablemarten",
    "vitalmouse",
]

CASES_TO_IGNORE = PROBLEMATIC_CASES + VALIDATION_CASES



PROP2UDF = {
    "application": "Sequencing Analysis",
    "bait_set": "Bait Set",
    "capture_kit": "Capture Library version",
    "comment": "Comment",
    "concentration": "Concentration (nM)",
    "concentration_sample": "Sample Conc.",
    "customer": "customer",
    "data_analysis": "Data Analysis",
    "elution_buffer": "Sample Buffer",
    "extraction_method": "Extraction method",
    "family_name": "familyID",
    "formalin_fixation_time": "Formalin Fixation Time",
    "index": "Index type",
    "index_number": "Index number",
    "lab_code": "Lab Code",
    "organism": "Strain",
    "organism_other": "Other species",
    "pool": "pool name",
    "post_formalin_fixation_time": "Post Formalin Fixation Time",
    "pre_processing_method": "Pre Processing Method",
    "priority": "priority",
    "quantity": "Quantity",
    "reference_genome": "Reference Genome Microbial",
    "region_code": "Region Code",
    "require_qcok": "Process only if QC OK",
    "rml_plate_name": "RML plate name",
    "selection_criteria": "Selection Criteria",
    "sex": "Gender",
    "source": "Source",
    "tissue_block_size": "Tissue Block Size",
    "target_reads": "Reads missing (M)",
    "tumour": "tumor",
    "tumour_purity": "tumour purity",
    "volume": "Volume (uL)",
    "well_position_rml": "RML well position",
    "verified_organism": "Verified organism",
    "sequencing_qc_pass": "Passed Sequencing QC",
}

MASTER_STEPS_UDFS = {
    "capture_kit_step": {
        "obsolete_CG002 - Hybridize Library  (SS XT)": "SureSelect capture library/libraries used",
        "Hybridize Library TWIST v1": "Bait Set",
        "Hybridize Library TWIST v2": "Bait Set",
    },
    "prep_method_step": {
        "CG002 - End repair Size selection A-tailing and Adapter ligation (TruSeq PCR-free DNA)": {
            "method_number": "Method document",
            "method_version": "Method document version",
        },
        "End repair Size selection A-tailing and Adapter ligation (TruSeq PCR-free DNA)": {
            "method_number": "Method document",
            "method_version": "Method document version",
        },
        "obsolete_CG002 - Hybridize Library  (SS XT)": {
            "method_number": "Method document",
            "method_version": "Method document versio",
        },
        "CG002 - Microbial Library Prep (Nextera)": {
            "method_number": "Method",
            "method_version": "Method Version",
        },
        "End-Repair and A-tailing TWIST v1": {
            "method_number": "Method document",
            "method_version": "Document version",
        },
        "KAPA Library Preparation TWIST v1": {
            "method_number": "Method document",
            "method_version": "Document version",
        },
    },
    "sequencing_method_step": {
        "CG002 - Cluster Generation (HiSeq X)": {
            "method_number": "Method",
            "method_version": "Version",
        },
        "CG002 - Cluster Generation (Illumina SBS)": {
            "method_number": "Method Document 1",
            "method_version": "Document 1 Version",
        },
        "Define Run Format and Calculate Volumes (Nova Seq)": {
            "method_number": "Method",
            "method_version": "Version",
        },
    },
    "delivery_method_step": {
        "CG002 - Delivery": {
            "method_number": "Method Document",
            "method_version": "Method Version",
        },
        "Delivery v1": {"method_number": "Method Document", "method_version": "Method Version"},
    },
}

PROCESSES = {"sequenced_date": "AUTOMATED - NovaSeq Run"}



"""Utility functions for the upload cli commands"""
import logging
from typing import List
import click

from cg.constants import Pipeline
from cg.store import models

LOG = logging.getLogger(__name__)


class LinkHelper:
    """Class that helps handle links"""

    @staticmethod
    def all_samples_are_non_tumour(links: List[models.FamilySample]) -> bool:
        """Return True if all samples are non tumour."""
        return all(not link.sample.is_tumour for link in links)

    @staticmethod
    def get_analysis_type_for_each_link(links: List[models.FamilySample]) -> list:
        """Return analysis type for each sample given by link list"""
        return [link.sample.application_version.application.analysis_type for link in links]


def suggest_cases_to_upload(context: click.Context, pipeline: Pipeline) -> None:
    LOG.warning("provide a case, suggestions:")
    records = context.obj["analysis_api"].status_db.analyses_to_upload(pipeline=pipeline)[:50]
    for case_obj in records:
        click.echo(case_obj)


def suggest_cases_delivery_report(context: click.Context, pipeline: Pipeline) -> None:
    LOG.error("provide a case, suggestions:")
    records = context.obj["analysis_api"].status_db.analyses_to_delivery_report(pipeline=pipeline)[
        :50
    ]
    for case_obj in records:
        click.echo(case_obj)



""" Module for retrieving FCs from backup """
import logging
import subprocess
import time

from cg.apps.pdc import PdcApi
from cg.store import Store, models

LOG = logging.getLogger(__name__)


class BackupApi:
    """ Class for retrieving FCs from backup """

    def __init__(self, status: Store, pdc_api: PdcApi, max_flowcells_on_disk: int, root_dir: str):
        self.status = status
        self.pdc = pdc_api
        self.max_flowcells_on_disk = max_flowcells_on_disk
        self.root_dir = root_dir

    def maximum_flowcells_ondisk(self) -> bool:
        """Check if there's too many flowcells already "ondisk"."""
        ondisk_flowcells = self.status.flowcells(status="ondisk").count()
        LOG.debug("ondisk flowcells: %s", ondisk_flowcells)
        return ondisk_flowcells > self.max_flowcells_on_disk

    def check_processing(self, max_processing_flowcells: int = 1) -> bool:
        """Check if the processing queue for flowcells is not full."""
        processing_flowcells = self.status.flowcells(status="processing").count()
        LOG.debug("processing flowcells: %s", processing_flowcells)
        return processing_flowcells < max_processing_flowcells

    def pop_flowcell(self, dry_run) -> models.Flowcell:
        """
        Get the top flowcell from the requested queue and update status to
        "processing".
        """
        flowcell_obj = self.status.flowcells(status="requested").first()
        if flowcell_obj is not None:
            flowcell_obj.status = "processing"
            if not dry_run:
                self.status.commit()
        return flowcell_obj

    def fetch_flowcell(self, flowcell_obj: models.Flowcell = None, dry_run: bool = False) -> float:
        """Start fetching a flowcell from backup if possible.

        1. The processing queue is not full
        2. The requested queue is not emtpy
        """
        if self.check_processing() is False:
            LOG.info("processing queue is full")
            return None

        if self.maximum_flowcells_ondisk() is True:
            LOG.info("maximum flowcells ondisk reached")
            return None

        if flowcell_obj:
            flowcell_obj.status = "processing"
            if not dry_run:
                self.status.commit()

        else:
            flowcell_obj = self.pop_flowcell(dry_run)
            if flowcell_obj is None:
                LOG.info("no flowcells requested")
                return None

        if not dry_run:
            LOG.info("%s: retrieving from PDC", flowcell_obj.name)

        start_time = time.time()

        try:
            self.pdc.retrieve_flowcell(
                flowcell_obj.name, flowcell_obj.sequencer_type, self.root_dir, dry_run
            )
            if not dry_run:
                flowcell_obj.status = "retrieved"
                self.status.commit()
                LOG.info('Status for flowcell %s set to "retrieved"', flowcell_obj.name)
        except subprocess.CalledProcessError as error:
            LOG.error("%s: retrieval failed", flowcell_obj.name)
            if not dry_run:
                flowcell_obj.status = "requested"
                self.status.commit()
            raise error

        end_time = time.time()
        return end_time - start_time



"""Tests for the scout upload API"""
from pathlib import Path

import pytest
import yaml
from cg.meta.upload.scout.mip_config_builder import MipConfigBuilder
from cg.meta.upload.scout.scoutapi import UploadScoutAPI
from cg.models.scout.scout_load_config import MipLoadConfig, ScoutLoadConfig


def test_unlinked_family_is_linked(mip_config_builder: MipConfigBuilder):
    """Test that is_family check fails when samples are not linked"""
    # GIVEN a upload scout api and case data for a case without linked individuals
    family_data: MipLoadConfig = MipLoadConfig(
        **{
            "samples": [
                {"sample_id": "ADM2", "father": "0", "mother": "0"},
                {"sample_id": "ADM3", "father": "0", "mother": "0"},
            ]
        }
    )
    # WHEN running the check if case is linked
    res = mip_config_builder.is_family_case(load_config=family_data)
    # THEN assert that the test returns False
    assert res is False


def test_family_is_linked(mip_config_builder: MipConfigBuilder):
    """Test that is_family returns true when samples are linked"""
    # GIVEN a upload scout api and case data for a linked case
    family_data: MipLoadConfig = MipLoadConfig(
        **{
            "samples": [
                {"sample_id": "ADM1", "father": "ADM2", "mother": "ADM3"},
                {"sample_id": "ADM2", "father": "0", "mother": "0"},
                {"sample_id": "ADM3", "father": "0", "mother": "0"},
            ]
        }
    )
    # WHEN running the check if case is linked
    res = mip_config_builder.is_family_case(family_data)
    # THEN assert that the test returns True
    assert res is True


def _file_exists_on_disk(file_path: Path):
    """Returns if the file exists on disk"""
    return file_path.exists()


def _file_is_yaml(file_path):
    """Returns if the file successfully was loaded as yaml"""
    data = None
    with open(file_path, "r") as stream:
        try:
            data = yaml.safe_load(stream)
        except yaml.YAMLError:
            return False
    return data is not None


def test_save_config_creates_file(
    upload_scout_api: UploadScoutAPI, mip_load_config: ScoutLoadConfig, tmp_file
):
    """"Tests that save config creates a file"""

    # GIVEN a scout_config object and a path to save it on

    # WHEN calling method to save config
    upload_scout_api.save_config_file(upload_config=mip_load_config, file_path=tmp_file)

    # THEN the config should exist on disk
    assert _file_exists_on_disk(tmp_file)


def test_save_config_creates_yaml(
    upload_scout_api: UploadScoutAPI, mip_load_config: ScoutLoadConfig, tmp_file
):
    """Tests that the file created by save_config_file create a yaml """

    # GIVEN a scout_config dict and a path to save it on

    # WHEN calling method to save config
    upload_scout_api.save_config_file(upload_config=mip_load_config, file_path=tmp_file)

    # THEN the should be of yaml type
    assert _file_is_yaml(tmp_file)


def test_add_scout_config_to_hk(upload_scout_api: UploadScoutAPI, tmp_file):
    """Test that scout load config is added to housekeeper"""
    # GIVEN a hk_mock and a file path to scout load config
    tag_name = UploadScoutAPI.get_load_config_tag()
    housekeeper_api = upload_scout_api.housekeeper
    # GIVEN a hk mock that does not return a config file
    housekeeper_api.add_missing_tag(tag_name)
    # WHEN adding the file path to hk_api

    upload_scout_api.add_scout_config_to_hk(config_file_path=tmp_file, case_id="dummy")
    # THEN assert that the file path is added to hk
    # file added to hk-db
    assert housekeeper_api.is_file_added() is True
    assert housekeeper_api.is_file_included() is True


def test_add_scout_config_to_hk_existing_files(upload_scout_api: UploadScoutAPI, tmp_file):
    """Test that scout config is not updated in housekeeper if it already exists"""
    # GIVEN a hk_mock with an scout upload file and a file path to scout load config
    housekeeper_api = upload_scout_api.housekeeper
    # GIVEN that there are files in the hk mock
    housekeeper_api.new_file(tmp_file.name)
    assert housekeeper_api.files
    # WHEN adding the file path to hk_api
    with pytest.raises(FileExistsError):
        # THEN assert File exists exception is raised
        upload_scout_api.add_scout_config_to_hk(config_file_path=tmp_file, case_id="dummy")



"""Code for uploading delivery report from the CLI"""
import logging
import sys
from pathlib import Path

import click

from cg.constants import Pipeline
from cg.exc import CgError, DeliveryReportError
from cg.meta.report.api import ReportAPI
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI

from .utils import suggest_cases_delivery_report

LOG = logging.getLogger(__name__)
SUCCESS = 0
FAIL = 1


@click.command("delivery-reports")
@click.option("-p", "--print", "print_console", is_flag=True, help="print list to console")
@click.option(
    "-f",
    "--force",
    "force_report",
    is_flag=True,
    help="overrule report validation",
)
@click.pass_context
def delivery_reports(context, print_console, force_report):
    """Generate delivery reports for all cases that need one"""

    if not context.obj.get("analysis_api"):
        context.obj["analysis_api"] = MipDNAAnalysisAPI(context.obj)
    analysis_api = context.obj["analysis_api"]

    click.echo(click.style("----------------- DELIVERY REPORTS ------------------------"))

    exit_code = SUCCESS
    for analysis_obj in analysis_api.status_db.analyses_to_delivery_report(
        pipeline=Pipeline.MIP_DNA
    ):
        case_id = analysis_obj.family.internal_id
        LOG.info("Uploading delivery report for case: %s", case_id)
        try:

            context.invoke(
                delivery_report,
                case_id=analysis_obj.family.internal_id,
                print_console=print_console,
                force_report=force_report,
            )
        except FileNotFoundError as error:
            LOG.error(
                "Missing file for delivery report creation for case: %s, %s",
                case_id,
                error,
            )
            exit_code = FAIL
        except DeliveryReportError as error:
            LOG.error(
                "Creation of delivery report failed for case: %s, %s",
                case_id,
                error.message,
            )
            exit_code = FAIL
        except CgError as error:
            LOG.error(
                "Uploading delivery report failed for case: %s, %s",
                case_id,
                error.message,
            )
            exit_code = FAIL
        except Exception:
            LOG.error(
                "Unspecified error when uploading delivery report for case: %s",
                case_id,
            )
            exit_code = FAIL
    sys.exit(exit_code)


@click.command("delivery-report")
@click.argument("case_id", required=False)
@click.option(
    "-p",
    "--print",
    "print_console",
    is_flag=True,
    help="print report to console",
)
@click.option(
    "-f",
    "--force",
    "force_report",
    is_flag=True,
    help="overrule report validation",
)
@click.option(
    "--analysis-started-at", help="Use the analysis started at (i.e.  '2020-05-28  12:00:46')"
)
@click.pass_context
def delivery_report(
    context: click.Context,
    case_id: str,
    print_console: bool,
    force_report: bool,
    analysis_started_at: str = None,
) -> None:
    """Generates a delivery report for a case and uploads it to housekeeper and scout

    The report contains data from several sources:

    status-db:
        family
        family.data_analysis        missing on most re-runs
        customer_name
        applications
        accredited
        panels
        samples
        sample.internal_id
        sample.status
        sample.ticket
        sample.million_read_pairs   for sequenced samples, from demux + ready made libraries (rml), not for external
        sample.prepared_at          not for rml and external
        sample.received_at
        sample.sequenced_at         for rml and in-house sequenced samples
        sample.delivered_at

    lims:
        sample.name
        sample.sex
        sample.source               missing on most re-runs
        sample.application
        sample.prep_method          not for rml or external
        sample.sequencing_method    for sequenced samples


    trailblazer:
        sample.mapped_reads
        sample.duplicates
        sample.analysis_sex
        pipeline_version
        genome_build

    chanjo:
        sample.target_coverage
        sample.target_completeness

    scout:
        panel-genes

    calculated:
        today
        sample.processing_time
        report_version

    """

    click.echo(click.style("----------------- DELIVERY_REPORT -------------"))

    analysis_api: MipDNAAnalysisAPI = context.obj["analysis_api"]
    report_api: ReportAPI = context.obj["report_api"]

    if not case_id or not analysis_api.status_db.family(internal_id=case_id):
        suggest_cases_delivery_report(context, pipeline=Pipeline.MIP_DNA)
        context.abort()

    if not analysis_started_at:
        analysis_started_at = analysis_api.status_db.family(case_id).analyses[0].started_at

    LOG.info("Using analysis started at: %s", analysis_started_at)

    if print_console:
        delivery_report_html = report_api.create_delivery_report(
            case_id, analysis_started_at, force_report
        )
        click.echo(delivery_report_html)
        return

    delivery_report_file = report_api.create_delivery_report_file(
        case_id,
        file_path=Path(analysis_api.root, case_id),
        accept_missing_data=force_report,
        analysis_date=analysis_started_at,
    )

    added_file = report_api.add_delivery_report_to_hk(
        delivery_report_file, analysis_api.housekeeper_api, case_id, analysis_started_at
    )

    if added_file:
        click.echo(click.style("Uploaded to housekeeper", fg="green"))
    else:
        click.echo(click.style("Already uploaded to housekeeper, skipping", fg="yellow"))

    context.invoke(delivery_report_to_scout, case_id=case_id)
    report_api.update_delivery_report_date(
        status_api=analysis_api.status_db, case_id=case_id, analysis_date=analysis_started_at
    )


@click.command("delivery-report-to-scout")
@click.argument("case_id", required=False)
@click.option(
    "-d",
    "--dry-run",
    "dry_run",
    is_flag=True,
    help="run command without uploading to scout",
)
@click.pass_context
def delivery_report_to_scout(context, case_id: str, dry_run: bool):
    """Fetches an delivery-report from housekeeper and uploads it to scout"""
    if not context.obj.get("analysis_api"):
        context.obj["analysis_api"] = MipDNAAnalysisAPI(context.obj)
    analysis_api = context.obj["analysis_api"]

    if not case_id:
        suggest_cases_delivery_report(context, pipeline=Pipeline.MIP_DNA)
        context.abort()

    uploaded_delivery_report_files = analysis_api.housekeeper_api.get_files(
        bundle=case_id,
        tags=["delivery-report"],
        version=analysis_api.housekeeper_api.last_version(case_id).id,
    )
    if uploaded_delivery_report_files.count() == 0:
        raise FileNotFoundError(f"No delivery report was found in housekeeper for {case_id}")

    report_path = uploaded_delivery_report_files[0].full_path

    LOG.info("uploading delivery report %s to scout for case: %s", report_path, case_id)
    if not dry_run:
        analysis_api.scout_api.upload_delivery_report(
            report_path=report_path, case_id=case_id, update=True
        )
    click.echo(click.style("uploaded to scout", fg="green"))



import logging

import click

from cg.cli.compress.helpers import get_fastq_individuals, update_compress_api
from cg.exc import CaseNotFoundError

LOG = logging.getLogger(__name__)


@click.command("sample")
@click.argument("sample-id", type=str)
@click.option("-d", "--dry-run", is_flag=True)
@click.pass_context
def store_sample(context, sample_id, dry_run):
    """Include links to decompressed FASTQ files belonging to this sample in housekeeper"""
    compress_api = context.obj["compress_api"]
    update_compress_api(compress_api, dry_run=dry_run)

    was_decompressed = compress_api.add_decompressed_fastq(sample_id)
    if was_decompressed is False:
        LOG.info(f"Skipping sample {sample_id}")
        return 0
    LOG.info(f"Stored fastq files for {sample_id}")
    return 1


@click.command("case")
@click.argument("case-id", type=str)
@click.option("-d", "--dry-run", is_flag=True)
@click.pass_context
def store_case(context, case_id, dry_run):
    """Include links to decompressed FASTQ files belonging to this case in housekeeper"""

    status_db = context.obj["status_db"]
    try:
        samples = get_fastq_individuals(status_db, case_id)
        stored_inds = 0
        for sample_id in samples:
            stored_count = context.invoke(store_sample, sample_id=sample_id, dry_run=dry_run)
            stored_inds += stored_count
    except CaseNotFoundError:
        return
    LOG.info(f"Stored fastq files for {stored_inds} samples")


@click.command("flowcell")
@click.argument("flowcell_id", type=str)
@click.option("-d", "--dry-run", is_flag=True)
@click.pass_context
def store_flowcell(context, flowcell_id, dry_run):
    """Include links to decompressed FASTQ files belonging to this flowcell in housekeeper"""

    status_db = context.obj["status_db"]
    samples = status_db.get_samples_from_flowcell(flowcell_id=flowcell_id)
    stored_inds = 0
    for sample in samples:
        stored_count = context.invoke(store_sample, sample_id=sample.internal_id, dry_run=dry_run)
        stored_inds += stored_count
    LOG.info(f"Stored fastq files for {stored_inds} samples")


@click.command("ticket")
@click.argument("ticket_id", type=int)
@click.option("-d", "--dry-run", is_flag=True)
@click.pass_context
def store_ticket(context, ticket_id, dry_run):
    """Include links to decompressed FASTQ files belonging to this ticket in housekeeper"""
    status_db = context.obj["status_db"]
    samples = status_db.get_samples_from_ticket(ticket_id=ticket_id)
    stored_inds = 0
    for sample in samples:
        stored_count = context.invoke(store_sample, sample_id=sample.internal_id, dry_run=dry_run)
        stored_inds += stored_count
    LOG.info(f"Stored fastq files for {stored_inds} samples")



import logging
from pathlib import Path
from typing import Optional

import click
from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.models.demultiplex.flowcell import Flowcell
from click import Context

LOG = logging.getLogger(__name__)


@click.command(name="all")
@click.argument("flowcells-directory", type=click.Path(exists=True, file_okay=False))
@click.option("--dry-run", is_flag=True)
@click.pass_context
def demultiplex_all(
    context: Context,
    flowcells_directory: click.Path,
    dry_run: bool,
):
    """Demultiplex all flowcells that are ready under the flowcells_directory"""
    flowcells_directory: Path = Path(str(flowcells_directory))
    demultiplex_api: DemultiplexingAPI = context.obj["demultiplex_api"]
    demultiplex_api.set_dry_run(dry_run=dry_run)
    for sub_dir in flowcells_directory.iterdir():
        if not sub_dir.is_dir():
            continue
        LOG.info("Found directory %s", sub_dir)
        flowcell_obj = Flowcell(flowcell_path=sub_dir)
        if not demultiplex_api.is_demultiplexing_possible(flowcell=flowcell_obj) and not dry_run:
            continue

        if not flowcell_obj.validate_sample_sheet():
            LOG.warning(
                "Malformed sample sheet. Run cg samplesheet validate %s",
                flowcell_obj.sample_sheet_path,
            )
            if not dry_run:
                continue
        demultiplex_api.start_demultiplexing(flowcell=flowcell_obj)


@click.command(name="flowcell")
@click.argument("flowcell-directory", type=click.Path(file_okay=False, exists=True))
@click.option("--out-directory")
@click.option("--dry-run", is_flag=True)
@click.pass_context
def demultiplex_flowcell(
    context: Context,
    flowcell_directory: click.Path,
    out_directory: Optional[str],
    dry_run: bool,
):
    """Demultiplex a flowcell on slurm using CG"""
    LOG.info("Running cg demultiplex flowcell")
    flowcell_directory: Path = Path(str(flowcell_directory))
    demultiplex_api: DemultiplexingAPI = context.obj["demultiplex_api"]
    if out_directory:
        out_directory: Path = Path(out_directory)
        LOG.info("Set out_dir to %s", out_directory)
        demultiplex_api.out_dir = out_directory
    demultiplex_api.set_dry_run(dry_run=dry_run)
    flowcell_obj = Flowcell(flowcell_path=flowcell_directory)

    if not demultiplex_api.is_demultiplexing_possible(flowcell=flowcell_obj) and not dry_run:
        LOG.warning("Can not start demultiplexing!")
        return

    if not flowcell_obj.validate_sample_sheet():
        LOG.warning(
            "Malformed sample sheet. Run cg samplesheet validate %s", flowcell_obj.sample_sheet_path
        )
        if not dry_run:
            raise click.Abort
    demultiplex_api.start_demultiplexing(flowcell=flowcell_obj)



"""Functions that deals with dummy samples"""
from cg.apps.lims.samplesheet import LimsFlowcellSample


def dummy_sample_name(sample_name: str) -> str:
    """Convert a string to a dummy sample name

    Replace space and parentheses with dashes
    """
    return sample_name.replace(" ", "-").replace("(", "-").replace(")", "-")


def dummy_sample(flowcell: str, dummy_index: str, lane: int, name: str) -> LimsFlowcellSample:
    """ Constructs and returns a dummy sample in novaseq sample sheet format"""
    return LimsFlowcellSample(
        flowcell_id=flowcell,
        lane=lane,
        sample_id=dummy_sample_name(sample_name=name),
        index=dummy_index,
        sample_name="indexcheck",
        project="indexcheck",
    )



"""Base command for trending"""

import logging
from pathlib import Path
from typing import Tuple, Any, Optional, List

import click

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import Pipeline
from cg.exc import AnalysisUploadError
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.store import Store, models

LOG = logging.getLogger(__name__)

VOGUE_VALID_BIOINFO = [str(Pipeline.MIP_DNA), str(Pipeline.BALSAMIC)]


@click.group()
@click.pass_context
def vogue(context):
    """Load trending data into trending database"""

    if not context.obj.get("analysis_api"):
        context.obj["analysis_api"] = MipDNAAnalysisAPI(context.obj)

    click.echo(click.style("----------------- TRENDING -----------------------"))


@vogue.command("genotype", short_help="Getting genotype data from the genotype database.")
@click.option(
    "-d", "--days", type=int, required="True", help="load X days old sampels from genotype to vogue"
)
@click.pass_context
def genotype(context, days: int):
    """Loading samples from the genotype database to the trending database"""

    if not context.obj.get("analysis_api"):
        context.obj["analysis_api"] = MipDNAAnalysisAPI(context.obj)
    analysis_api = context.obj["analysis_api"]

    click.echo(click.style("----------------- GENOTYPE -----------------------"))

    analysis_api.upload_vogue_api.load_genotype(days=days)


@vogue.command("apptags", short_help="Getting application tags to the trending database.")
@click.pass_context
def apptags(context):
    """Loading apptags from status db to the trending database"""

    if not context.obj.get("analysis_api"):
        context.obj["analysis_api"] = MipDNAAnalysisAPI(context.obj)
    analysis_api = context.obj["analysis_api"]

    click.echo(click.style("----------------- APPLICATION TAGS -----------------------"))

    analysis_api.upload_vogue_api.load_apptags()


@vogue.command("flowcells", short_help="Getting flowcell data from the lims.")
@click.option(
    "-d", "--days", type=int, required="True", help="load X days old runs from lims to vogue"
)
@click.pass_context
def flowcells(context, days: int):
    """Loading runs from lims to the trending database"""

    if not context.obj.get("analysis_api"):
        context.obj["analysis_api"] = MipDNAAnalysisAPI(context.obj)
    analysis_api = context.obj["analysis_api"]

    LOG.info("----------------- FLOWCELLS -----------------------")

    analysis_api.vogue_api.load_flowcells(days=days)


@vogue.command("samples", short_help="Getting sample data from lims.")
@click.option(
    "-d", "--days", type=int, required="True", help="load X days old sampels from lims to vogue"
)
@click.pass_context
def samples(context, days: int):
    """Loading samples from lims to the trending database"""

    if not context.obj.get("analysis_api"):
        context.obj["analysis_api"] = MipDNAAnalysisAPI(context.obj)
    analysis_api = context.obj["analysis_api"]

    LOG.info("----------------- SAMPLES -----------------------")

    analysis_api.vogue_api.load_samples(days=days)


@vogue.command("reagent-labels", short_help="Getting reagent_label data from lims.")
@click.option(
    "-d", "--days", type=int, required=True, help="load X days old sampels from lims to vogue"
)
@click.pass_context
def reagent_labels(context, days: int):
    """Loading reagent_labels from lims to the trending database"""

    if not context.obj.get("analysis_api"):
        context.obj["analysis_api"] = MipDNAAnalysisAPI(context.obj)
    analysis_api = context.obj["analysis_api"]

    LOG.info("----------------- REAGENT LABELS -----------------------")

    analysis_api.vogue_api.load_reagent_labels(days=days)


@vogue.command("bioinfo", short_help="Load bioinfo results into vogue")
@click.option(
    "-c",
    "--case-name",
    required=True,
    help="Case name or project name for which the analysis results will load",
)
@click.option(
    "--cleanup/--no-cleanup", default=False, help="Cleanup processed case data while loading"
)
@click.option(
    "-t",
    "--target-load",
    default="all",
    type=click.Choice(["all", "raw", "process"]),
    show_default=True,
    help=(
        "Final target to load bioinfo data."
        "Target orders are: all, raw, process."
        "Selecting all, will load raw, process, and sample."
        "Selecting process, will upload raw and process."
        "Selecting raw, will only load raw data."
    ),
)
@click.option("--dry/--no-dry", default=False, help="Dry run...")
@click.pass_context
def bioinfo(context, case_name, cleanup, target_load, dry):
    """Load bioinfo case results to the trending database"""

    if not context.obj.get("analysis_api"):
        context.obj["analysis_api"] = MipDNAAnalysisAPI(context.obj)
    analysis_api = context.obj["analysis_api"]

    click.echo(click.style("----------------- BIOINFO -----------------------"))

    load_bioinfo_raw_inputs = dict()

    # Probably get samples for a case_name through statusdb api
    load_bioinfo_raw_inputs["samples"] = _get_samples(analysis_api.status_db, case_name)

    # Probably get analysis result file through housekeeper ai
    load_bioinfo_raw_inputs["analysis_result_file"] = _get_multiqc_latest_file(
        analysis_api.housekeeper_api, case_name
    )

    # Probably get analysis_type [multiqc or microsalt or all] from cli
    # This might automated to some extend by checking if input multiqc json.
    # This tells us how the result was generated. If it is multiqc it will try to validate keys with
    # an actual model.
    load_bioinfo_raw_inputs["analysis_type"] = "multiqc"

    # case_name is the input
    load_bioinfo_raw_inputs["analysis_case_name"] = case_name

    # Get case_analysis_type from cli a free text for an entry in trending database
    load_bioinfo_raw_inputs["case_analysis_type"] = "multiqc"

    # Get workflow_name and workflow_version
    workflow_name, workflow_version = _get_analysis_workflow_details(
        analysis_api.status_db, case_name
    )
    if workflow_name not in VOGUE_VALID_BIOINFO:
        raise AnalysisUploadError(f"Case upload failed: {case_name}. Reason: Bad workflow name.")
    load_bioinfo_raw_inputs["analysis_workflow_name"] = workflow_name
    load_bioinfo_raw_inputs["analysis_workflow_version"] = workflow_version

    if dry:
        click.echo(click.style("----------------- DRY RUN -----------------------"))

    if target_load in ("raw", "all"):
        click.echo(click.style("----------------- UPLOAD UNPROCESSED -----------------------"))
        if not dry:
            analysis_api.upload_vogue_api.load_bioinfo_raw(load_bioinfo_raw_inputs)

    if target_load in ("process", "all"):
        click.echo(click.style("----------------- PROCESS CASE -----------------------"))
        if not dry:
            analysis_api.upload_vogue_api.load_bioinfo_process(load_bioinfo_raw_inputs, cleanup)
        click.echo(click.style("----------------- PROCESS SAMPLE -----------------------"))
        if not dry:
            analysis_api.upload_vogue_api.load_bioinfo_sample(load_bioinfo_raw_inputs)


@vogue.command("bioinfo-all", short_help="Load all bioinfo results into vogue")
@click.option("--dry/--no-dry", is_flag=True, help="Dry run...")
@click.pass_context
def bioinfo_all(context, dry):
    """Load all cases with recent analysis and a multiqc-json to the trending database."""

    if not context.obj.get("analysis_api"):
        context.obj["analysis_api"] = MipDNAAnalysisAPI(context.obj)
    analysis_api = context.obj["analysis_api"]

    cases = analysis_api.status_db.families()
    for case in cases:
        case_name = case.internal_id
        version_obj = analysis_api.housekeeper_api.last_version(case_name)
        if not version_obj:
            continue

        # confirm multiqc.json exists
        multiqc_file_obj = analysis_api.housekeeper_api.get_files(
            bundle=case_name, tags=["multiqc-json"], version=version_obj.id
        )
        if len(list(multiqc_file_obj)) == 0:
            continue

        # confirm that file exists
        existing_multiqc_file = multiqc_file_obj[0].full_path
        if not Path(existing_multiqc_file).exists():
            continue

        LOG.info("Found multiqc for %s, %s", case_name, existing_multiqc_file)
        try:
            context.invoke(bioinfo, case_name=case_name, cleanup=True, target_load="all", dry=dry)
        except AnalysisUploadError:
            LOG.error("Case upload failed: %s", case_name, exc_info=True)


def _get_multiqc_latest_file(hk_api: HousekeeperAPI, case_name: str) -> str:
    """Get latest multiqc_data.json path for a case_name
    Args:
        case_name(str): onemite
    Returns:
        multiqc_data_path(str): /path/to/multiqc.json
    """
    version_obj = hk_api.last_version(case_name)
    multiqc_json_file = hk_api.get_files(
        bundle=case_name, tags=["multiqc-json"], version=version_obj.id
    )

    if len(list(multiqc_json_file)) == 0:
        raise FileNotFoundError(f"No multiqc.json was found in housekeeper for {case_name}")

    return multiqc_json_file[0].full_path


def _get_samples(store: Store, case_name: str) -> str:
    """Get a sample string for case_name
    Args:
        case_name(str): onemite
    Returns:
        sample_names(str): ACC12345,ACC45679
    """

    link_objs: List[models.FamilySample] = store.family(case_name).links
    sample_ids = {link_obj.sample.internal_id for link_obj in link_objs}
    return ",".join(sample_ids)


def _get_analysis_workflow_details(status_api: Store, case_name: str) -> Tuple[Any, Optional[Any]]:
    """Get lowercase workflow name for a case_name
    Args:
        case_name(str): onemite
    Returns:
        workflow_name(str): balsamic
        workflow_version(str): v3.14.15
    """
    # Workflow that generated these results
    case_obj = status_api.family(case_name)
    workflow_name = None
    workflow_version = None
    if case_obj.analyses:
        workflow_name = case_obj.analyses[0].pipeline
        workflow_version = case_obj.analyses[0].pipeline_version

    return workflow_name.lower(), workflow_version



import csv
import datetime as dt
import logging
import os
import shutil
from abc import ABC
from pathlib import Path
from subprocess import CalledProcessError
from typing import List

from alchy import Query

from cg.constants import Pipeline
from cg.exc import CgError
from cg.meta.workflow.analysis import AnalysisAPI
from cg.store import models
from cg.utils import Process

LOG = logging.getLogger(__name__)


class FluffyAnalysisAPI(AnalysisAPI):
    def __init__(
        self,
        config: dict,
        pipeline: Pipeline = Pipeline.FLUFFY,
    ):
        self.root_dir = Path(config["fluffy"]["root_dir"])
        self.fluffy_config = Path(config["fluffy"]["config_path"])
        super().__init__(
            pipeline,
            config,
        )

    @property
    def threshold_reads(self):
        return False

    @property
    def process(self) -> Process:
        if not self._process:
            self._process = Process(binary=self.config["fluffy"]["binary_path"])
        return self._process

    def get_samplesheet_path(self, case_id: str) -> Path:
        """
        Location in case folder where samplesheet is expected to be stored. Samplesheet is used as a config
        required to run Fluffy
        """
        return Path(self.root_dir, case_id, "SampleSheet.csv")

    def get_workdir_path(self, case_id: str) -> Path:
        """
        Location in case folder where all sub-folders for each sample containing fastq files are to e saved
        """
        return Path(self.root_dir, case_id, "fastq")

    def get_fastq_path(self, case_id: str, sample_id: str) -> Path:
        """
        Location in working directory to which fastq files are saved for each sample separately
        """
        return Path(self.get_workdir_path(case_id), sample_id)

    def get_output_path(self, case_id: str) -> Path:
        """
        Location in case directory where Fluffy outputs will be stored
        """
        return Path(self.root_dir, case_id, "output")

    def get_deliverables_file_path(self, case_id: str) -> Path:
        """
        Location in working directory where deliverables file will be stored upon completion of analysis.
        Deliverables file is used to communicate paths and tag definitions for files in a finished analysis
        """
        return Path(self.get_output_path(case_id), "deliverables.yaml")

    def get_trailblazer_config_path(self, case_id: str) -> Path:
        """
        Location in working directory where SLURM job id file is to be stored.
        This file contains SLURM ID of jobs associated with current analysis ,
        is used as a config to be submitted to Trailblazer and track job progress in SLURM
        """
        return Path(self.get_output_path(case_id), "sacct", "submitted_jobs.yaml")

    def get_analysis_finish_path(self, case_id: str) -> Path:
        return Path(self.get_output_path(case_id), "analysis_finished")

    def link_fastq_files(self, case_id: str, dry_run: bool = False) -> None:
        """
        Links fastq files from Housekeeper to case working directory
        """
        case_obj: models.Family = self.status_db.family(case_id)
        workdir_path = self.get_workdir_path(case_id=case_id)
        if workdir_path.exists() and not dry_run:
            LOG.info("Fastq directory exists, removing and re-linking files!")
            shutil.rmtree(workdir_path, ignore_errors=True)
        workdir_path.mkdir(parents=True, exist_ok=True)
        for familysample in case_obj.links:
            sample_id = familysample.sample.internal_id
            files: Query = self.housekeeper_api.files(bundle=sample_id, tags=["fastq"])
            sample_path: Path = self.get_fastq_path(case_id=case_id, sample_id=sample_id)
            for file in files:
                if not dry_run:
                    Path.mkdir(sample_path, exist_ok=True, parents=True)
                    Path(sample_path / Path(file.full_path).name).symlink_to(file.full_path)
                LOG.info(f"Linking {file.full_path} to {sample_path / Path(file.full_path).name}")

    def get_concentrations_from_lims(self, sample_id: str) -> str:
        """Get sample concentration from LIMS"""
        return self.lims_api.get_sample_attribute(lims_id=sample_id, key="concentration_sample")

    def add_concentrations_to_samplesheet(
        self, samplesheet_housekeeper_path: Path, samplesheet_workdir_path: Path
    ) -> None:
        """
        Reads the fluffy samplesheet *.csv file as found in Housekeeper.
        Edits column 'SampleName' to include customer name for sample.
        Appends a column 'Library_nM' at the end of the csv with concentration values extracted from LIMS
        """
        csv_reader = csv.reader(open(samplesheet_housekeeper_path, "r"))
        csv_writer = csv.writer(open(samplesheet_workdir_path, "w"))

        sample_name_index = None
        sample_id_index = None
        for row in csv_reader:
            if "SampleID" in row and not sample_id_index:
                sample_name_index = row.index("SampleName")
                sample_id_index = row.index("SampleID")
                row.append("Library_nM")
                csv_writer.writerow(row)
                continue
            if sample_id_index:
                sample_id = row[sample_id_index]
                row[sample_name_index] = self.get_sample_name_from_lims_id(lims_id=sample_id)
                row.append(str(self.get_concentrations_from_lims(sample_id=sample_id)))
                csv_writer.writerow(row)

    def get_samplesheet_housekeeper_path(self, flowcell_name: str) -> Path:
        """
        Returns the path to original samplesheet file that is added to Housekeeper
        """
        samplesheet_query: list = self.housekeeper_api.files(
            bundle=flowcell_name, tags=["samplesheet"]
        ).all()
        if not samplesheet_query:
            LOG.error(
                "Samplesheet file for flowcell %s could not be found in Housekeeper!", flowcell_name
            )
            raise CgError
        return Path(samplesheet_query[0].full_path)

    def make_samplesheet(self, case_id: str, dry_run: bool) -> None:
        """
        Create SampleSheet.csv file in working directory and add desired values to the file
        """
        case_obj: models.Family = self.status_db.family(case_id)
        flowcell_name: str = case_obj.links[0].sample.flowcells[0].name
        samplesheet_housekeeper_path = self.get_samplesheet_housekeeper_path(
            flowcell_name=flowcell_name
        )
        samplesheet_workdir_path = Path(self.get_samplesheet_path(case_id=case_id))
        LOG.info(
            "Writing modified csv from %s to %s",
            samplesheet_housekeeper_path,
            samplesheet_workdir_path,
        )
        if not dry_run:
            Path(self.root_dir, case_id).mkdir(parents=True, exist_ok=True)
            self.add_concentrations_to_samplesheet(
                samplesheet_housekeeper_path=samplesheet_housekeeper_path,
                samplesheet_workdir_path=samplesheet_workdir_path,
            )

    def run_fluffy(self, case_id: str, dry_run: bool) -> None:
        """
        Call fluffy with the configured command-line arguments
        """
        output_path: Path = self.get_output_path(case_id=case_id)
        if output_path.exists():
            LOG.info("Old working directory found, cleaning!")
            if not dry_run:
                shutil.rmtree(output_path, ignore_errors=True)
        command_args = [
            "--config",
            self.fluffy_config.as_posix(),
            "--sample",
            self.get_samplesheet_path(case_id=case_id).as_posix(),
            "--project",
            self.get_workdir_path(case_id=case_id).as_posix(),
            "--out",
            self.get_output_path(case_id=case_id).as_posix(),
            "analyse",
        ]
        self.process.run_command(command_args, dry_run=dry_run)

    def get_cases_to_store(self) -> List[models.Family]:
        """Retrieve a list of cases where analysis finished successfully,
        and is ready to be stored in Housekeeper"""
        return [
            case_object
            for case_object in self.get_running_cases()
            if Path(self.get_analysis_finish_path(case_id=case_object.internal_id)).exists()
        ]



from datetime import datetime
from typing import Optional

from pydantic import BaseModel, validator
from typing_extensions import Literal

SEX_MAP = {"male": "M", "female": "F"}


class Udf(BaseModel):
    application: str
    capture_kit: Optional[str]
    comment: Optional[str]
    concentration: Optional[str]
    concentration_sample: Optional[str]
    customer: str
    data_analysis: Optional[str]
    data_delivery: Optional[str]
    elution_buffer: Optional[str]
    extraction_method: Optional[str]
    family_name: str = "NA"
    formalin_fixation_time: Optional[str]
    index: Optional[str]
    index_number: Optional[str]
    lab_code: Optional[str]
    organism: Optional[str]
    organism_other: Optional[str]
    pool: Optional[str]
    post_formalin_fixation_time: Optional[str]
    pre_processing_method: Optional[str]
    priority: str = "standard"
    quantity: Optional[str]
    reference_genome: Optional[str]
    region_code: Optional[str]
    require_qcok: bool = False
    rml_plate_name: Optional[str]
    selection_criteria: Optional[str]
    sex: Literal["M", "F", "unknown"] = "unknown"
    source: str = "NA"
    tissue_block_size: Optional[str]
    tumour: bool = False
    tumour_purity: Optional[str]
    volume: Optional[str]
    well_position_rml: Optional[str]
    verified_organism: Optional[bool]

    @validator("sex", pre=True)
    def validate_sex(cls, value: str):
        return SEX_MAP.get(value, "unknown")


class LimsSample(BaseModel):
    name: str
    container: str = "Tube"
    container_name: Optional[str]
    well_position: Optional[str]
    index_sequence: Optional[str]
    udfs: Optional[Udf]



"""Tests for FASTQ part of meta compress api"""
import logging


def test_compress_case_fastq_one_sample(populated_compress_fastq_api, sample, caplog):
    """Test to compress all FASTQ files for a sample"""
    caplog.set_level(logging.DEBUG)
    compress_api = populated_compress_fastq_api
    compress_api.crunchy_api.set_dry_run(True)
    # GIVEN a populated compress api

    # WHEN Compressing the bam files for the case
    res = compress_api.compress_fastq(sample)

    # THEN assert compression succeded
    assert res is True
    # THEN assert that the correct information is communicated
    assert "Compressing" in caplog.text
    # THEN assert that the correct information is communicated
    assert "to SPRING format" in caplog.text


def test_compress_fastq_compression_done(
    populated_compress_fastq_api, compression_object, sample, caplog
):
    """Test to compress all FASTQ files for a sample when compression is already completed

    The program should not compress any files since the compression is done for a sample
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN a populated compress api
    compress_api = populated_compress_fastq_api
    # GIVEN that the spring file exists (compression is done)
    compression_object.spring_path.touch()

    # WHEN Compressing the bam files for the case
    res = compress_api.compress_fastq(sample)

    # THEN assert compression succeded
    assert res is False
    # THEN assert that the correct information is communicated
    assert f"FASTQ to SPRING not possible for {sample}" in caplog.text


def test_compress_case_fastq_compression_pending(
    populated_compress_fastq_api, sample, compression_object, caplog
):
    """Test to compress all FASTQ files for a sample when compression is pending

    The program should not compress any files since the compression is pending for a sample
    """
    caplog.set_level(logging.DEBUG)
    compress_api = populated_compress_fastq_api
    # GIVEN that the pending flag exists
    compression_object.pending_path.touch()

    # WHEN compressing the FASTQ files for the case
    res = compress_api.compress_fastq(sample)

    # THEN assert compression returns False
    assert res is False
    # THEN assert that the correct information is communicated
    assert f"FASTQ to SPRING not possible for {sample}" in caplog.text



"""Functions that deals with modifications of the indexes"""
import csv
import logging
from typing import Dict, List, Set, Tuple

from cg.apps.lims.samplesheet import LimsFlowcellSample
from cg.resources import valid_indexes_path
from pydantic import BaseModel

LOG = logging.getLogger(__name__)
NEW_CONTROL_SOFTWARE_VERSION = "1.7.0"
NEW_REAGENT_KIT_VERSION = 1.5
DNA_COMPLEMENTS = {"A": "T", "C": "G", "G": "C", "T": "A"}
REAGENT_KIT_PARAMETER_TO_VERSION = {"1": 1.0, "3": 1.5}


def index_exists(index: str, indexes: Set[str]) -> bool:
    """ Determines if a index is already present in the existing indexes """
    return any(existing_index.startswith(index) for existing_index in indexes)


def get_indexes_by_lane(samples: List[LimsFlowcellSample]) -> Dict[int, Set[str]]:
    """Group the indexes from samples by lane"""
    indexes_by_lane = {}
    for sample in samples:
        lane: int = sample.lane
        if lane not in indexes_by_lane:
            indexes_by_lane[lane] = set()
        indexes_by_lane[lane].add(sample.index)
    return indexes_by_lane


class Index(BaseModel):
    name: str
    sequence: str


def get_valid_indexes(dual_indexes_only: bool = True) -> List[Index]:
    LOG.info("Fetch valid indexes from %s", valid_indexes_path)
    indexes: List[Index] = []
    with open(valid_indexes_path, "r") as csv_file:
        indexes_csv = csv.reader(csv_file)
        for row in indexes_csv:
            index_name = row[0]
            index_sequence = row[1]
            if dual_indexes_only and not is_dual_index(index_sequence):
                continue
            indexes.append(Index(name=index_name, sequence=index_sequence))
    return indexes


def get_reagent_kit_version(reagent_kit_version: str) -> float:
    """ Derives the reagent kit version from the run parameters """
    LOG.info("Converting reagent kit parameter %s to version", reagent_kit_version)
    if reagent_kit_version not in REAGENT_KIT_PARAMETER_TO_VERSION:
        raise SyntaxError(f"Unknown reagent kit version {reagent_kit_version}")

    return REAGENT_KIT_PARAMETER_TO_VERSION[reagent_kit_version]


def is_reverse_complement(control_software_version: str, reagent_kit_version_string: str) -> bool:
    """If the run used the new NovaSeq control software version (NEW_CONTROL_SOFTWARE_VERSION)
    and the new reagent kit version (NEW_REAGENT_KIT_VERSION) the second index should be the
    reverse complement
    """
    LOG.info("Check if run is reverse complement")
    if control_software_version != NEW_CONTROL_SOFTWARE_VERSION:
        LOG.warning(
            "Old software version %s, no need for reverse complement", control_software_version
        )
        return False
    reagent_kit_version: float = get_reagent_kit_version(reagent_kit_version_string)
    if reagent_kit_version != NEW_REAGENT_KIT_VERSION:
        LOG.warning(
            "Reagent kit version %s does not does not need reverse complement",
            reagent_kit_version,
        )
        return False
    LOG.info("Run is reverse complement")
    return True


def get_reverse_complement_dna_seq(dna: str) -> str:
    """ Generates the reverse complement of a DNA sequence"""
    LOG.debug("Reverse complement string %s", dna)

    return "".join(DNA_COMPLEMENTS[base] for base in reversed(dna))


def pad_index_one(index_string: str) -> str:
    """Adds bases 'AT' to index one"""
    return index_string + "AT"


def pad_index_two(index_string: str, reverse_complement: bool) -> str:
    """Adds bases to index two depending on if it should be reverse complement or not"""
    if reverse_complement:
        return "AC" + index_string
    return index_string + "AC"


def adapt_indexes(
    samples: List[LimsFlowcellSample],
    control_software_version: str,
    reagent_kit_version: str,
    expected_index_length: int,
) -> None:
    """Adapts the indexes: pads all indexes so that all indexes have a length equal to the
    number  of index reads, and takes the reverse complement of index 2 in case of the new
    novaseq software control version (1.7) in combination with the new reagent kit
    (version 1.5)
    """
    LOG.info("Fix so that all indexes are on the correct format")
    reverse_complement: bool = is_reverse_complement(
        control_software_version=control_software_version,
        reagent_kit_version_string=reagent_kit_version,
    )
    for sample in samples:
        index1, index2 = sample.index.split("-")
        index_length = len(index1)
        if expected_index_length == 10 and index_length == 8:
            LOG.debug("Padding indexes")
            index1 = pad_index_one(index_string=index1)
            index2 = pad_index_two(index_string=index2, reverse_complement=reverse_complement)
        if reverse_complement:
            index2 = get_reverse_complement_dna_seq(index2)
        sample.index = index1
        sample.index2 = index2


def is_dual_index(index: str) -> bool:
    """ Determines if an index in the raw sample sheet is dual index or not """
    return "-" in index



"""Constants for cg"""
from cgmodels.cg.constants import Pipeline, StrEnum

CONTAINER_OPTIONS = ("Tube", "96 well plate")

CAPTUREKIT_OPTIONS = (
    "Agilent Sureselect CRE",
    "Agilent Sureselect V5",
    "SureSelect Focused Exome",
    "Twist_Target_hg19.bed",
    "other",
)

CAPTUREKIT_CANCER_OPTIONS = (
    "GIcfDNA",
    "GMCKsolid",
    "GMSmyeloid",
    "LymphoMATIC",
    "other (specify in comment field)",
)

COMBOS = {
    "DSD": ("DSD", "HYP", "SEXDIF", "SEXDET"),
    "CM": ("CNM", "CM"),
    "Horsel": ("Horsel", "141217", "141201"),
}
COLLABORATORS = ("cust000", "cust002", "cust003", "cust004", "cust042")

DEFAULT_CAPTURE_KIT = "twistexomerefseq_9.1_hg19_design.bed"

CASE_ACTIONS = ("analyze", "running", "hold")

FLOWCELL_STATUS = ("ondisk", "removed", "requested", "processing", "retrieved")


class DataDelivery(StrEnum):
    ANALYSIS_FILES: str = "analysis"
    ANALYSIS_BAM_FILES: str = "analysis-bam"
    FASTQ: str = "fastq"
    NIPT_VIEWER: str = "nipt-viewer"
    FASTQ_QC: str = "fastq_qc"
    SCOUT: str = "scout"


PREP_CATEGORIES = ("cov", "mic", "rml", "tgs", "wes", "wgs", "wts")

SEX_OPTIONS = ("male", "female", "unknown")

STATUS_OPTIONS = ("affected", "unaffected", "unknown")
ANALYSIS_TYPES = ["tumor_wgs", "tumor_normal_wgs", "tumor_panel", "tumor_normal_panel"]



"""Mock a Process"""

import copy
import logging
from subprocess import CalledProcessError
from typing import List

from cg.constants import RETURN_SUCCESS
from cg.utils.commands import Process

LOG = logging.getLogger(__name__)


class ProcessMock(Process):
    """Mock the utils.commands.Process class"""

    def __init__(
        self,
        binary: str = None,
        config: str = None,
        config_parameter: str = "--config",
        environment: str = None,
    ):
        """
        Args:
            binary(str): Path to binary for the process to use
            config(str): Path to config if used by process
            environment(str): Activate conda environment before executing binary
        """
        super(ProcessMock, self).__init__(
            binary=binary, config=config, config_parameter=config_parameter, environment=environment
        )
        self.config_parameter = config_parameter
        LOG.debug("Initialising Process with binary: %s", self.binary)
        self.base_call: List = []
        if self.binary:
            self.base_call.append(self.binary)
            if self.environment:
                LOG.debug("Activating environment with: %s", self.environment)
                self.base_call.insert(0, f"source activate {self.environment};")
            if self.config:
                self.base_call.extend([self.config_parameter, self.config])
            LOG.debug("Use base call %s", self.base_call)

        self._exit_code: int = RETURN_SUCCESS

    def run_command(self, parameters: list = None, dry_run: bool = False) -> int:
        """Execute a command in the shell.
        If environment is supplied - shell=True has to be supplied to enable passing as a string for executing multiple
         commands

        Args:
            parameters(list):
            dry_run(bool): Print command instead of executing it
        Return(int): Return code from called process

        """
        command = copy.deepcopy(self.base_call)
        if parameters:
            command.extend(parameters)

        LOG.info("Running command %s", " ".join(command))
        if dry_run:
            return RETURN_SUCCESS

        if self._exit_code != RETURN_SUCCESS:
            LOG.critical("Call %s exit with a non zero exit code", command)
            LOG.critical(self.stderr)
            raise CalledProcessError(self._exit_code, command)

        return RETURN_SUCCESS

    def set_stdout(self, text: str):
        """Mock the stdout"""
        self._stdout = text

    def set_stderr(self, text: str):
        """Mock the stderr"""
        self._stderr = text

    def set_exit_code(self, code: int):
        """Mock the stderr"""
        self._exit_code = code

    def set_binary(self, binary_path: str):
        """Set the binary path"""
        self.binary = binary_path
        self.base_call.append(self.binary)
        if self.environment:
            LOG.debug("Activating environment with: %s", self.environment)
            self.base_call.insert(0, f"source activate {self.environment};")
        if self.config:
            self.base_call.extend([self.config_parameter, self.config])
        LOG.debug("Use base call %s", self.base_call)



from enum import Enum
from typing import List, Optional

from cg.constants import DataDelivery, Pipeline
from pydantic import BaseModel, constr


class SexEnum(str, Enum):
    male = "male"
    female = "female"
    other = "unknown"


class PriorityEnum(str, Enum):
    research = "research"
    standard = "standard"
    priority = "priority"
    express = "express"
    clinical_trials = "clinical trials"


class ContainerEnum(str, Enum):
    tube = "Tube"
    plate = "96 well plate"


class CaptureKitEnum(str, Enum):
    agilent_sureselect_cre = "Agilent Sureselect CRE"
    agilent_sureselect_v5 = "Agilent Sureselect V5"
    sureselect_focused_exome = "SureSelect Focused Exome"
    twist_target_hg19_bed = "Twist_Target_hg19.bed"
    other = "other"


class StatusEnum(str, Enum):
    affected = "affected"
    unaffected = "unaffected"
    unknown = "unknown"


NAME_PATTERN = r"^[A-Za-z0-9-]*$"


class OrderSample(BaseModel):
    age_at_sampling: Optional[str]
    application: str
    capture_kit: Optional[CaptureKitEnum]
    case_id: str
    cohorts: Optional[List[str]]
    comment: Optional[str]
    concentration: Optional[float]
    concentration_sample: Optional[float]
    container: Optional[ContainerEnum]
    container_name: Optional[str]
    customer: Optional[str]
    custom_index: Optional[str]
    data_analysis: Pipeline
    data_delivery: DataDelivery
    elution_buffer: Optional[str]
    extraction_method: Optional[str]
    family_name: Optional[str]
    father: Optional[str]
    formalin_fixation_time: Optional[int]
    from_sample: Optional[str]
    index: str
    index_number: Optional[str]
    index_sequence: Optional[str]
    internal_id: Optional[str]
    mother: Optional[str]
    name: constr(regex=NAME_PATTERN)
    organism: Optional[str]
    organism_other: Optional[str]
    panels: Optional[List[str]]
    phenotype_terms: Optional[List[str]]
    pool: Optional[str]
    post_formalin_fixation_time: Optional[int]
    priority: PriorityEnum = PriorityEnum.standard
    quantity: Optional[int]
    reagent_label: Optional[str]
    reference_genome: Optional[str]
    require_qcok: bool = False
    rml_plate_name: Optional[str]
    sex: SexEnum = SexEnum.other
    source: Optional[str]
    status: StatusEnum = StatusEnum.unknown
    synopsis: Optional[str]
    time_point: Optional[int]
    tissue_block_size: Optional[str]
    tumour: bool = False
    tumour_purity: Optional[int]
    verified_organism: Optional[bool]
    volume: Optional[str]
    well_position: Optional[str]
    well_position_rml: Optional[str]



"""This script tests the cli mip store functions"""
import logging

from click.testing import CliRunner

from cg.cli.workflow.mip.store import analysis
from cg.constants import EXIT_FAIL, EXIT_SUCCESS


def test_store_no_config(cli_runner: CliRunner, dna_mip_context: dict, caplog):
    """Test the function to check for config files"""

    caplog.set_level(logging.INFO)
    # WHEN we run store analysis on no config
    result = cli_runner.invoke(analysis, obj=dna_mip_context)
    # THEN We should be informed to enter a config file
    assert "Provide a config file." in caplog.text
    # THEN the exit code should be EXIT_FAIL
    assert result.exit_code == EXIT_FAIL


def test_store_analysis_exception(
    cli_runner: CliRunner, dna_mip_context: dict, mip_configs: dict, caplog
):
    """Test that the analysis function enters the exception clause"""
    with caplog.at_level("ERROR", "INFO"):
        # WHEN we run store on a case without deliverables
        result = cli_runner.invoke(analysis, [str(mip_configs["purplesnail"])], obj=dna_mip_context)
        # THEN we should be informed that mandatory files are missing
        assert "Mandatory files are missing" in caplog.text
        # THEN we should not include any files in housekeeper
        assert "included files in Housekeeper" not in caplog.text
        # THEN the exit code should be EXIT_FAIL
        assert result.exit_code == EXIT_FAIL


def test_store_analysis(
    cli_runner: CliRunner,
    mip_configs: dict,
    dna_mip_context: dict,
    mip_qc_sample_info: dict,
    caplog,
):
    """Test if store completed stores a completed sample"""

    with caplog.at_level("INFO"):
        # WHEN we run store on a config
        result = cli_runner.invoke(analysis, [str(mip_configs["yellowhog"])], obj=dna_mip_context)
        # THEN we should store the files in housekeeper
        assert "new bundle added: yellowhog" in caplog.text
        assert "Included files in Housekeeper" in caplog.text
        # THEN the exit code should be EXIT_SUCCESS
        assert result.exit_code == EXIT_SUCCESS



"""Fixtures for the model tests"""
import gzip
import io
from pathlib import Path

import pytest


@pytest.fixture(name="content")
def fixture_content() -> str:
    """Return some content for a file"""
    _content = (
        "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt"
        " ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ull"
        "amco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehende"
        "rit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaec"
        "at cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum."
    )
    return _content


@pytest.fixture(name="non_existing_file_path")
def fixture_non_existing_file_path(project_dir: Path) -> Path:
    """Return the path to a non existing file"""
    return project_dir / "a_file.txt"


@pytest.fixture(name="non_existing_gzipped_file_path")
def fixture_non_existing_gzipped_file_path(non_existing_file_path: Path) -> Path:
    """Return the path to a non existing file with gzipped ending"""
    return non_existing_file_path.with_suffix(".gz")


@pytest.fixture(name="filled_file")
def fixture_filled_file(non_existing_file_path: Path, content: str) -> Path:
    """Return the path to a existing file with some content"""
    with open(non_existing_file_path, "w") as outfile:
        outfile.write(content)
    return non_existing_file_path


@pytest.fixture(name="filled_gzip_file")
def fixture_filled_gzip_file(non_existing_gzipped_file_path: Path, content: str) -> Path:
    """Return the path to a existing file with some content that is gzipped"""
    with gzip.open(non_existing_gzipped_file_path, "wb") as outfile:
        with io.TextIOWrapper(outfile, encoding="utf-8") as enc:
            enc.write(content)
    return non_existing_gzipped_file_path


@pytest.fixture(scope="function", name="gzipped_empty_file")
def fixture_gzipped_empty_file(fastq_dir: Path) -> Path:
    """Return a the path to a file that is gzipped without any content

    This file will not have the size = 0 due to gzip header and footer
    """

    return fastq_dir / "dummy_run_R1_001.fastq.gz"



import logging

import click
from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.cli.demultiplex.demux import demultiplex_all, demultiplex_flowcell
from cg.cli.demultiplex.sample_sheet import sample_sheet_commands
from click import Context

LOG = logging.getLogger(__name__)


@click.group()
@click.pass_context
def demultiplex(context: Context):
    """Command group for the demultiplex commands"""
    context.obj["demultiplex_api"] = DemultiplexingAPI(config=context.obj)


demultiplex.add_command(demultiplex_flowcell)
demultiplex.add_command(demultiplex_all)
demultiplex.add_command(sample_sheet_commands)



"""CLI for delivering files with CG"""
import copy
import logging
from pathlib import Path
from typing import List

import click

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants.delivery import PIPELINE_ANALYSIS_OPTIONS, PIPELINE_ANALYSIS_TAG_MAP
from cg.meta.deliver import DeliverAPI
from cg.store import Store
from cg.store.models import Family

LOG = logging.getLogger(__name__)


@click.group()
@click.pass_context
def deliver(context):
    """Deliver files with CG."""
    LOG.info("Running CG deliver")
    context.obj["status_db"] = Store(context.obj["database"])
    context.obj["housekeeper_api"] = HousekeeperAPI(context.obj)


@click.command(name="analysis")
@click.option("-c", "--case-id", help="Deliver the files for a specific case")
@click.option(
    "-t", "--ticket-id", type=int, help="Deliver the files for ALL cases connected to a ticket"
)
@click.option("-d", "--delivery-type", type=click.Choice(PIPELINE_ANALYSIS_OPTIONS), required=True)
@click.option("--dry-run", is_flag=True)
@click.pass_context
def deliver_analysis(context, case_id: str, ticket_id: int, delivery_type: str, dry_run: bool):
    """Deliver analysis files to customer inbox

    Files can be delivered either on case level or for all cases connected to a ticket.
    Any of those needs to be specified.
    """
    if not (case_id or ticket_id):
        LOG.info("Please provide a case-id or ticket-id")
        return

    inbox = context.obj.get("delivery_path")
    if not inbox:
        LOG.info("Please specify the root path for where files should be delivered")
        return

    status_db: Store = context.obj["status_db"]
    deliver_api = DeliverAPI(
        store=status_db,
        hk_api=context.obj["housekeeper_api"],
        case_tags=PIPELINE_ANALYSIS_TAG_MAP[delivery_type]["case_tags"],
        sample_tags=PIPELINE_ANALYSIS_TAG_MAP[delivery_type]["sample_tags"],
        project_base_path=Path(inbox),
        delivery_type=delivery_type,
    )
    deliver_api.set_dry_run(dry_run)
    if case_id:
        case_obj = status_db.family(case_id)
        if not case_obj:
            LOG.warning("Could not find case %s", case_id)
            return
        cases = [case_obj]
    else:
        cases: List[Family] = status_db.get_cases_from_ticket(ticket_id=ticket_id).all()
        if not cases:
            LOG.warning("Could not find cases for ticket_id %s", ticket_id)
            return

    for case_obj in cases:
        deliver_api.deliver_files(case_obj=case_obj)


@click.command(name="old-analysis")
@click.option("-c", "--case-id", help="Deliver the files for a specific case")
@click.option(
    "-t", "--ticket-id", type=int, help="Deliver the files for ALL cases connected to a ticket"
)
@click.option("-d", "--delivery-type", type=click.Choice(PIPELINE_ANALYSIS_OPTIONS), required=True)
@click.option("--dry-run", is_flag=True)
@click.pass_context
def deliver_old_analysis(context, case_id: str, ticket_id: int, delivery_type: str, dry_run: bool):
    """Deliver old analysis files to customer inbox

    Files can be delivered either on case level or for all cases connected to a ticket.
    Any of those needs to be specified.
    """

    BALSAMIC_ANALYSIS_ONLY_CASE_TAGS = [
        {"vcf-snv-clinical"},
        {"vcf-snv-clinical-index"},
        {"vcf-pass"},
        {"vcf-sv-clinical"},
        {"vcf-sv-clinical-index"},
        {"cnvkit", "visualization"},
        {"cnv-scatter", "scatter"},
        {"cnvkit", "visualization", "diagram"},
        {"cnv-diagram", "diagram"},
        {"cnr"},
        {"multiqc-html"},
    ]

    BALSAMIC_ANALYSIS_CASE_TAGS = copy.deepcopy(BALSAMIC_ANALYSIS_ONLY_CASE_TAGS)
    BALSAMIC_ANALYSIS_CASE_TAGS.extend(
        [
            {"cram", "normal"},
            {"cram-index"},
            {"cram", "tumor"},
            {"cram-index", "tumor"},
            {"cram", "tumor-cram"},
            {"cram", "tumor-cram-index"},
            {"cram", "normal-cram"},
            {"cram", "normal-cram-index"},
        ]
    )

    BALSAMIC_ANALYSIS_SAMPLE_TAGS = [
        {"cram", "normal"},
        {"cram-index"},
        {"cram", "tumor"},
        {"cram-index", "tumor"},
        {"bam"},
        {"bam-index"},
        {"bam", "tumor-bam"},
        {"bam", "tumor-bam-index"},
        {"bam", "normal-bam"},
        {"bam", "normal-bam-index"},
        {"cram", "tumor-cram"},
        {"cram", "tumor-cram-index"},
        {"cram", "normal-cram"},
        {"cram", "normal-cram-index"},
    ]

    BALSAMIC_QC_CASE_TAGS = [
        {"multiqc-html"},
    ]
    BALSAMIC_QC_SAMPLE_TAGS = [{"fastq"}, {"read1"}, {"read2"}]

    OLD_PIPELINE_ANALYSIS_TAG_MAP = {
        "balsamic": {
            "case_tags": BALSAMIC_ANALYSIS_CASE_TAGS,
            "sample_tags": BALSAMIC_ANALYSIS_SAMPLE_TAGS,
        },
        "balsamic-analysis": {"case_tags": BALSAMIC_ANALYSIS_ONLY_CASE_TAGS, "sample_tags": []},
        "balsamic-qc": {
            "case_tags": BALSAMIC_QC_CASE_TAGS,
            "sample_tags": BALSAMIC_QC_SAMPLE_TAGS,
        },
    }
    if not (case_id or ticket_id):
        LOG.info("Please provide a case-id or ticket-id")
        return

    inbox = context.obj.get("delivery_path")
    if not inbox:
        LOG.info("Please specify the root path for where files should be delivered")
        return

    status_db: Store = context.obj["status_db"]
    deliver_api = DeliverAPI(
        store=status_db,
        hk_api=context.obj["housekeeper_api"],
        case_tags=OLD_PIPELINE_ANALYSIS_TAG_MAP[delivery_type]["case_tags"],
        sample_tags=OLD_PIPELINE_ANALYSIS_TAG_MAP[delivery_type]["sample_tags"],
        project_base_path=Path(inbox),
    )
    deliver_api.set_dry_run(dry_run)
    if case_id:
        case_obj = status_db.family(case_id)
        if not case_obj:
            LOG.warning("Could not find case %s", case_id)
            return
        cases = [case_obj]
    else:
        cases: List[Family] = status_db.get_cases_from_ticket(ticket_id=ticket_id).all()
        if not cases:
            LOG.warning("Could not find cases for ticket_id %s", ticket_id)
            return

    for case_obj in cases:
        deliver_api.deliver_files(case_obj=case_obj)


deliver.add_command(deliver_analysis)
deliver.add_command(deliver_old_analysis)



"""Cli commands for importing data into the status database"""
import logging
import getpass
from pathlib import Path

import click
from cg.constants import PREP_CATEGORIES
from cg.store import Store
from cg.store.api.import_func import (
    import_application_versions,
    import_applications,
    import_apptags,
)
from cg.utils.click.EnumChoice import EnumChoice

LOG = logging.getLogger(__name__)


@click.group("import")
@click.pass_context
def import_cmd(context):
    """Import information into the database."""
    context.obj["status_db"] = Store(context.obj["database"])


@import_cmd.command("application")
@click.argument("excel_path")
@click.argument("signature", required=False)
@click.option(
    "-d", "--dry-run", "dry_run", is_flag=True, help="Test run, no changes to the " "database"
)
@click.option("-s", "--sheet-name", help="Sheet name in workbook")
@click.pass_context
def application(context, excel_path, signature, sheet_name, dry_run):
    """Import new applications to status-db"""

    if not signature:
        signature = getpass.getuser()

    import_applications(context.obj["status_db"], excel_path, signature, dry_run, sheet_name)


@import_cmd.command("application-version")
@click.argument("excel_path")
@click.argument("signature", required=False)
@click.option(
    "-d", "--dry-run", "dry_run", is_flag=True, help="Test run, no changes to the " "database"
)
@click.option(
    "--skip-missing", "skip_missing", is_flag=True, help="continue despite missing " "applications"
)
@click.pass_context
def application_version(context, excel_path, signature, dry_run, skip_missing):
    """Import new application versions to status-db"""

    if not signature:
        signature = getpass.getuser()

    import_application_versions(
        context.obj["status_db"], excel_path, signature, dry_run, skip_missing
    )


@import_cmd.command("apptag")
@click.argument("excel_path", type=click.types.Path(exists=True, file_okay=True, dir_okay=False))
@click.argument("prep-category", type=click.types.Choice(PREP_CATEGORIES))
@click.argument("signature", required=False, type=click.types.STRING)
@click.option(
    "-s",
    "--sheet-name",
    default="Drop down list",
    help="(optional) name of sheet where to find the applications",
)
@click.option(
    "-c",
    "--tag-column",
    default=2,
    help="(optional) zero-based column where to find the applications",
    type=click.types.INT,
)
@click.option(
    "-r",
    "--tag-row",
    default=0,
    help="(optional) zero-based row where to start looking for the applications",
    type=click.types.INT,
)
@click.option(
    "-a", "--activate", is_flag=True, help="Activate archived tags found in the " "orderform"
)
@click.option("-i", "--inactivate", is_flag=True, help="Inactivate tags not found in the orderform")
@click.pass_context
def apptag(
    context: click.Context,
    excel_path: Path,
    prep_category: str,
    signature: str,
    sheet_name: str,
    tag_column: int,
    tag_row: int,
    activate: bool,
    inactivate: bool,
):
    """
    Syncs all applications from the specified excel file
    Args:
        :param inactivate:          Inactivate tags not found in the orderform
        :param activate:            Activate archived tags found in the orderform
        :param sheet_name:          (optional) name of sheet where the applications can be found
        :param tag_column:          (optional) zero-based column where the application tags can be
        found
        :param tag_row:             (optional) zero-based row where the application tags can be
        found
        :param prep_category:       prep_category to sync
        :param excel_path:          Path to orderform excel file
        :param signature:           Signature of user running the script
    """

    if not signature:
        signature = getpass.getuser()

    import_apptags(
        store=context.obj["status_db"],
        excel_path=excel_path,
        prep_category=prep_category,
        signature=signature,
        sheet_name=sheet_name,
        tag_column=tag_column,
        tag_row=tag_row,
        activate=activate,
        inactivate=inactivate,
    )



"""Tests for running the demultiplex flowcell command"""
import logging
from pathlib import Path
from typing import Dict

from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.cli.demultiplex.demux import demultiplex_flowcell
from cg.models.demultiplex.flowcell import Flowcell
from click import testing


def test_demultiplex_flowcell_dry_run(
    cli_runner: testing.CliRunner,
    demultiplex_ready_flowcell: Path,
    demultiplex_context: Dict[str, DemultiplexingAPI],
    caplog,
):
    caplog.set_level(logging.INFO)
    # GIVEN that all files are present for demultiplexing
    flowcell: Flowcell = Flowcell(demultiplex_ready_flowcell)
    # GIVEN a out dir that does not exist
    demux_api: DemultiplexingAPI = demultiplex_context["demultiplex_api"]
    assert demux_api.is_demultiplexing_possible(flowcell=flowcell)
    demux_dir: Path = demux_api.flowcell_out_dir_path(flowcell)
    unaligned_dir: Path = demux_dir / "Unaligned"
    assert demux_dir.exists() is False
    assert unaligned_dir.exists() is False

    # WHEN starting demultiplexing from the CLI with dry run flag
    result: testing.Result = cli_runner.invoke(
        demultiplex_flowcell,
        [str(demultiplex_ready_flowcell), "--dry-run"],
        obj=demultiplex_context,
    )
    # THEN assert the command exits without problems
    assert result.exit_code == 0
    # THEN assert no results folder was created since it is run in dry run mode
    assert demux_dir.exists() is False
    assert unaligned_dir.exists() is False


def test_demultiplex_flowcell(
    cli_runner: testing.CliRunner,
    demultiplex_ready_flowcell: Path,
    demultiplex_context: Dict[str, DemultiplexingAPI],
    caplog,
):
    caplog.set_level(logging.INFO)
    # GIVEN that all files are present for demultiplexing
    flowcell: Flowcell = Flowcell(demultiplex_ready_flowcell)
    # GIVEN a out dir that does not exist
    demux_api: DemultiplexingAPI = demultiplex_context["demultiplex_api"]
    demux_dir: Path = demux_api.flowcell_out_dir_path(flowcell)
    unaligned_dir: Path = demux_dir / "Unaligned"
    assert demux_api.is_demultiplexing_possible(flowcell=flowcell)
    assert demux_dir.exists() is False
    assert unaligned_dir.exists() is False

    # WHEN starting demultiplexing from the CLI with dry run flag
    result: testing.Result = cli_runner.invoke(
        demultiplex_flowcell,
        [str(demultiplex_ready_flowcell)],
        obj=demultiplex_context,
    )
    # THEN assert the command exits without problems
    assert result.exit_code == 0
    # THEN assert the results folder was created
    assert demux_dir.exists()
    assert unaligned_dir.exists()
    # THEN assert that the sbatch script was created
    assert demux_api.demultiplex_sbatch_path(flowcell).exists()


def test_start_demultiplexing_when_already_completed(
    cli_runner: testing.CliRunner,
    demultiplex_ready_flowcell: Path,
    demultiplex_context: Dict[str, DemultiplexingAPI],
    caplog,
):
    caplog.set_level(logging.DEBUG)
    # GIVEN that all files are present for demultiplexing
    flowcell: Flowcell = Flowcell(demultiplex_ready_flowcell)
    demux_api: DemultiplexingAPI = demultiplex_context["demultiplex_api"]
    # GIVEN that demultiplexing has started
    flowcell.demultiplexing_started_path.touch()
    # GIVEN a out dir that exist
    demux_api.flowcell_out_dir_path(flowcell).mkdir(parents=True)
    # GIVEN that demultiplexing is completed
    demux_api.demultiplexing_completed_path(flowcell=flowcell).touch()

    # WHEN starting demultiplexing from the CLI
    result: testing.Result = cli_runner.invoke(
        demultiplex_flowcell,
        [str(demultiplex_ready_flowcell)],
        obj=demultiplex_context,
    )
    # THEN assert the command exits without problems
    assert result.exit_code == 0
    # THEN assert it was communicated that demultiplexing was completed
    assert f"Demultiplexing is already completed for flowcell {flowcell.flowcell_id}"



""" Module to group PDC related commands """

import logging
import subprocess

LOG = logging.getLogger(__name__)


class PdcApi:
    """ Group PDC related commands """

    @classmethod
    def retrieve_flowcell(
        cls, flowcell_id: str, sequencer_type: str, root_dir: str, dry: bool = False
    ) -> str:
        """Fetch a flowcell back from the backup solution."""
        server = {
            "hiseqga": "thalamus",
            "hiseqx": "hasta",
            "novaseq": "hasta",
        }.get(sequencer_type)
        path = root_dir[sequencer_type]

        if server is None:
            raise ValueError(f"{sequencer_type}: invalid sequencer type")

        bash_command = f"bash SCRIPTS/retrieve_run_nas.bash {flowcell_id} {server} {path}"
        command = ["ssh", "nas-9.scilifelab.se", bash_command]
        LOG.info(" ".join(command))
        if not dry:
            subprocess.check_call(command)



"""Code for uploading observations data via CLI"""
import logging

import click

from cg.apps.loqus import LoqusdbAPI
from cg.constants import Pipeline
from cg.exc import DuplicateRecordError, DuplicateSampleError
from cg.meta.upload.observations import UploadObservationsAPI

from .utils import LinkHelper
from ...meta.workflow.mip_dna import MipDNAAnalysisAPI

LOG = logging.getLogger(__name__)


@click.command()
@click.option("-c", "--case_id", help="internal case id, leave empty to process all")
@click.option("-l", "--case-limit", type=int, help="maximum number of cases to upload")
@click.option("--dry-run", is_flag=True, help="only print cases to be processed")
@click.pass_context
def observations(context, case_id, case_limit, dry_run):
    """Upload observations from an analysis to LoqusDB."""

    click.echo(click.style("----------------- OBSERVATIONS ----------------"))

    if not context.obj.get("analysis_api"):
        context.obj["analysis_api"] = MipDNAAnalysisAPI(context.obj)
    analysis_api = context.obj["analysis_api"]

    loqus_apis = {
        "wgs": LoqusdbAPI(context.obj),
        "wes": LoqusdbAPI(context.obj, analysis_type="wes"),
    }

    if case_id:
        families_to_upload = [analysis_api.status_db.family(case_id)]
    else:
        families_to_upload = analysis_api.status_db.observations_to_upload()

    nr_uploaded = 0
    for case_obj in families_to_upload:

        if case_limit is not None and nr_uploaded >= case_limit:
            LOG.info("Uploaded %d cases, observations upload will now stop", nr_uploaded)
            return

        if not case_obj.customer.loqus_upload:
            LOG.info(
                "%s: %s not whitelisted for upload to loqusdb. Skipping!",
                case_obj.internal_id,
                case_obj.customer.internal_id,
            )
            continue

        if case_obj.data_analysis.lower() != str(Pipeline.MIP_DNA):
            LOG.info("%s: has non-MIP data_analysis. Skipping!", case_obj.internal_id)
            continue

        if not LinkHelper.all_samples_are_non_tumour(case_obj.links):
            LOG.info("%s: has tumour samples. Skipping!", case_obj.internal_id)
            continue

        analysis_list = LinkHelper.get_analysis_type_for_each_link(case_obj.links)
        if len(set(analysis_list)) != 1 or analysis_list[0] not in (
            "wes",
            "wgs",
        ):
            LOG.info(
                "%s: Undetermined analysis type (wes or wgs) or mixed analyses. Skipping!",
                case_obj.internal_id,
            )
            continue

        analysis_type = analysis_list[0]

        if dry_run:
            LOG.info("%s: Would upload observations", case_obj.internal_id)
            continue

        api = UploadObservationsAPI(
            analysis_api.status_db, analysis_api.housekeeper_api, loqus_apis[analysis_type]
        )

        try:
            api.process(case_obj.analyses[0])
            LOG.info("%s: observations uploaded!", case_obj.internal_id)
            nr_uploaded += 1
        except (DuplicateRecordError, DuplicateSampleError) as error:
            LOG.info("%s: skipping observations upload: %s", case_obj.internal_id, error.message)
        except FileNotFoundError as error:
            LOG.info("%s: skipping observations upload: %s", case_obj.internal_id, error)



"""Test the cli for uploading delivery reports"""

import logging
from datetime import datetime

from cg.cli.upload.base import delivery_report
from cg.constants import Pipeline

EXIT_CODE_SUCCESS = 0


def test_no_parameters(upload_context, cli_runner, caplog):
    # GIVEN we have a report_api set up
    assert upload_context.get("report_api")

    # WHEN calling delivery_report without parameters
    with caplog.at_level(logging.INFO):
        result = cli_runner.invoke(delivery_report, [], obj=upload_context)

    # THEN it should fail on missing case_id
    assert result.exit_code != EXIT_CODE_SUCCESS
    assert "provide a case, suggestions:" in caplog.text


def test_analysis_started_at(upload_context, cli_runner, caplog, helpers):

    # GIVEN a correct case_id and a correct date
    analysis = helpers.add_analysis(
        upload_context["analysis_api"].status_db,
        started_at=datetime.now(),
        pipeline=Pipeline.MIP_DNA,
    )
    case_id = analysis.family.internal_id
    a_date = analysis.started_at
    assert a_date

    # WHEN calling delivery_report with ok date parameter
    with caplog.at_level(logging.INFO):
        result = cli_runner.invoke(
            delivery_report,
            [case_id, "--analysis-started-at", a_date],
            obj=upload_context,
        )

    # THEN it should contain the date in the logged info
    assert str(a_date) in caplog.text


def test_analysis_without_started_at(upload_context, cli_runner, caplog, helpers):

    # GIVEN a correct case_id and a correct date
    analysis = helpers.add_analysis(
        upload_context["analysis_api"].status_db,
        started_at=datetime.now(),
        pipeline=Pipeline.MIP_DNA,
    )
    case_id = analysis.family.internal_id
    a_date = analysis.started_at
    assert a_date

    caplog.set_level(logging.INFO)

    # WHEN calling delivery_report without date parameter
    cli_runner.invoke(
        delivery_report,
        [case_id],
        obj=upload_context,
    )

    # THEN it should contain the date in the logged info
    assert "Using analysis started at: " in caplog.text



from typing import Optional

from pydantic import BaseModel
from typing_extensions import Literal


class Sbatch(BaseModel):
    job_name: str
    account: str
    number_tasks: int
    memory: int
    log_dir: str
    email: str
    hours: int
    minutes: str = "00"
    priority: Literal["high", "normal", "low"] = "low"
    commands: str
    error: Optional[str]



"""CLI function to compress FASTQ files into SPRING archives"""

import datetime as dt
import logging

import click

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants.compression import CASES_TO_IGNORE
from cg.exc import CaseNotFoundError
from cg.meta.compress import CompressAPI
from cg.store import Store

from .helpers import correct_spring_paths, get_fastq_individuals, update_compress_api

LOG = logging.getLogger(__name__)

# There is a list of problematic cases that we should skip


@click.command("fastq")
@click.option("-c", "--case-id", type=str)
@click.option("-n", "--number-of-conversions", default=5, type=int, show_default=True)
@click.option("-t", "--ntasks", default=12, show_default=True, help="Number of tasks for slurm job")
@click.option("-m", "--mem", default=50, show_default=True, help="Memory for slurm job")
@click.option(
    "-b",
    "--days-back",
    default=60,
    show_default=True,
    help="Threshold for how long ago family was created",
)
@click.option("-d", "--dry-run", is_flag=True)
@click.pass_context
def fastq_cmd(context, case_id, number_of_conversions, ntasks, mem, days_back, dry_run):
    """ Find cases with FASTQ files and compress into SPRING """
    LOG.info("Running compress FASTQ")
    compress_api: CompressAPI = context.obj["compress_api"]
    store: Store = context.obj["status_db"]
    update_compress_api(compress_api, dry_run=dry_run, ntasks=ntasks, mem=mem)

    if case_id:
        case_obj = store.family(case_id)
        if not case_obj:
            LOG.warning("Could not find case %s", case_id)
            return
        cases = [case_obj]
    else:
        date_threshold = dt.datetime.now() - dt.timedelta(days=days_back)
        cases = compress_api.get_cases_to_compress(store, date_threshold=date_threshold)

    case_conversion_count = 0
    ind_conversion_count = 0
    for case in cases:
        # Keeps track on if all samples in a case have been converted
        case_converted = True
        if case_conversion_count >= number_of_conversions:
            break
        internal_id = case.internal_id
        if internal_id in CASES_TO_IGNORE:
            LOG.info("Skipping case %s", internal_id)
            continue

        LOG.info("Searching for FASTQ files in case %s", internal_id)
        for link_obj in case.links:
            sample_id = link_obj.sample.internal_id
            case_converted = compress_api.compress_fastq(sample_id)
            if case_converted is False:
                LOG.info("skipping individual %s", sample_id)
                continue
            ind_conversion_count += 1
        if case_converted:
            case_conversion_count += 1

    LOG.info(
        "%s Individuals in %s (completed) cases where compressed",
        ind_conversion_count,
        case_conversion_count,
    )


@click.command("fastq")
@click.option("-c", "--case-id")
@click.option(
    "-b",
    "--days-back",
    default=60,
    show_default=True,
    help="Threshold for how long ago family was created",
)
@click.option("-d", "--dry-run", is_flag=True)
@click.pass_context
def clean_fastq(context, case_id, days_back, dry_run):
    """Remove compressed FASTQ files, and update links in housekeeper to SPRING files"""
    LOG.info("Running compress clean FASTQ")
    compress_api: CompressAPI = context.obj["compress_api"]
    store: Store = context.obj["status_db"]
    update_compress_api(compress_api, dry_run=dry_run)

    if case_id:
        case_obj = store.family(case_id)
        if not case_obj:
            LOG.warning("Could not find case %s", case_id)
            return
        cases = [case_obj]
    else:
        date_threshold = dt.datetime.now() - dt.timedelta(days=days_back)
        cases = compress_api.get_cases_to_compress(store, date_threshold=date_threshold)

    cleaned_inds = 0
    for case_obj in cases:
        if case_obj.internal_id in CASES_TO_IGNORE:
            continue
        samples = get_fastq_individuals(store=store, case_id=case_obj.internal_id)
        for sample_id in samples:
            res = compress_api.clean_fastq(sample_id)
            if res is False:
                LOG.info("Skipping individual %s", sample_id)
                continue
            cleaned_inds += 1

    LOG.info("Cleaned fastqs in %s individuals", cleaned_inds)


@click.command("fix-spring")
@click.option("-b", "--bundle-name")
@click.option("-d", "--dry-run", is_flag=True)
@click.pass_context
def fix_spring(context, bundle_name, dry_run):
    """Check if bundle(s) have non existing SPRING files and correct these"""
    LOG.info("Running compress clean FASTQ")
    compress_api = context.obj["compress_api"]
    update_compress_api(compress_api, dry_run=dry_run)
    hk_api: HousekeeperAPI = compress_api.hk_api
    correct_spring_paths(hk_api=hk_api, bundle_name=bundle_name, dry_run=dry_run)


@click.command("sample")
@click.argument("sample-id", type=str)
@click.option("-d", "--dry-run", is_flag=True)
@click.pass_context
def decompress_sample(context, sample_id, dry_run):
    compress_api: CompressAPI = context.obj["compress_api"]
    update_compress_api(compress_api, dry_run=dry_run)

    was_decompressed = compress_api.decompress_spring(sample_id)
    if was_decompressed is False:
        LOG.info(f"Skipping sample {sample_id}")
        return 0
    LOG.info(f"Decompressed sample {sample_id}")
    return 1


@click.command("case")
@click.argument("case-id", type=str)
@click.option("-d", "--dry-run", is_flag=True)
@click.pass_context
def decompress_case(context, case_id, dry_run):
    """Decompress SPRING file, and include links to FASTQ files in housekeeper"""

    store: Store = context.obj["status_db"]
    try:
        samples = get_fastq_individuals(store, case_id)
        decompressed_inds = 0
        for sample_id in samples:
            decompressed_count = context.invoke(
                decompress_sample, sample_id=sample_id, dry_run=dry_run
            )
            decompressed_inds += decompressed_count
    except CaseNotFoundError:
        return
    LOG.info(f"Decompressed spring archives in {decompressed_inds} samples")


@click.command("flowcell")
@click.argument("flowcell_id", type=str)
@click.option("-d", "--dry-run", is_flag=True)
@click.pass_context
def decompress_flowcell(context, flowcell_id, dry_run):
    """Decompress SPRING file, and include links to FASTQ files in housekeeper"""

    store: Store = context.obj["status_db"]
    samples = store.get_samples_from_flowcell(flowcell_id=flowcell_id)
    decompressed_inds = 0
    for sample in samples:
        decompressed_count = context.invoke(
            decompress_sample, sample_id=sample.internal_id, dry_run=dry_run
        )
        decompressed_inds += decompressed_count
    LOG.info(f"Decompressed spring archives in {decompressed_inds} samples")


@click.command("ticket")
@click.argument("ticket_id", type=int)
@click.option("-d", "--dry-run", is_flag=True)
@click.pass_context
def decompress_ticket(context, ticket_id, dry_run):
    """Decompress SPRING file, and include links to FASTQ files in housekeeper"""
    store: Store = context.obj["status_db"]
    samples = store.get_samples_from_ticket(ticket_id=ticket_id)
    decompressed_inds = 0
    for sample in samples:
        decompressed_count = context.invoke(
            decompress_sample, sample_id=sample.internal_id, dry_run=dry_run
        )
        decompressed_inds += decompressed_count
    LOG.info(f"Decompressed spring archives in {decompressed_inds} samples")



import json
from pathlib import Path
from typing import List

import pytest
from cg.apps.demultiplex.sample_sheet.index import Index
from cg.apps.demultiplex.sample_sheet.novaseq_sample_sheet import SampleSheetCreator
from cg.apps.lims.samplesheet import LimsFlowcellSample
from cg.models.demultiplex.run_parameters import RunParameters


@pytest.fixture(name="flowcell_name")
def fixture_flowcell_name() -> str:
    return "HVKJCDRXX"


@pytest.fixture(name="index_obj")
def fixture_index_obj() -> Index:
    return Index(name="C07 - UDI0051", sequence="AACAGGTT-ATACCAAG")


@pytest.fixture(name="demultiplex_fixtures")
def fixture_demultiplex_fixtures(apps_dir: Path) -> Path:
    """Return the path to the demultiplex fixtures"""
    return apps_dir / "demultiplexing"


@pytest.fixture(name="raw_samples_dir")
def fixture_raw_samples_dir(demultiplex_fixtures: Path) -> Path:
    """Return the path to the raw samples fixtures"""
    return demultiplex_fixtures / "raw_lims_samples"


@pytest.fixture(name="novaseq_dir")
def fixture_novaseq_dir(demultiplex_fixtures: Path) -> Path:
    """Return the path to the novaseq demultiplex fixtures"""
    return demultiplex_fixtures / "201203_A00689_0200_AHVKJCDRXX"


@pytest.fixture(name="hiseq_dir")
def fixture_hiseq_dir(demultiplex_fixtures: Path) -> Path:
    """Return the path to the hiseq demultiplex fixtures"""
    return demultiplex_fixtures / "hiseq_run"


@pytest.fixture(name="unknown_run_parameters")
def fixture_unknown_run_parameters(demultiplex_fixtures: Path) -> Path:
    """Return the path to a file with hiseq run parameters"""
    return demultiplex_fixtures / "unknown_run_parameters.xml"


@pytest.fixture(name="run_parameters_missing_flowcell_type")
def fixture_run_parameters_missing_flowcell_type(demultiplex_fixtures: Path) -> Path:
    """Return the path to a file with hiseq run parameters"""
    return demultiplex_fixtures / "runParameters_missing_flowcell_run_field.xml"


@pytest.fixture(name="hiseq_run_parameters")
def fixture_hiseq_run_parameters(hiseq_dir: Path) -> Path:
    """Return the path to a file with hiseq run parameters"""
    return hiseq_dir / "runParameters.xml"


@pytest.fixture(name="novaseq_run_parameters")
def fixture_novaseq_run_parameters(novaseq_dir: Path) -> Path:
    """Return the path to a file with hiseq run parameters"""
    return novaseq_dir / "RunParameters.xml"


@pytest.fixture(name="raw_lims_sample")
def fixture_raw_lims_sample(flowcell_name: str) -> LimsFlowcellSample:
    """Return a raw lims sample"""
    sample = {
        "flowcell_id": flowcell_name,
        "lane": 1,
        "sample_id": "ACC7628A20",
        "sample_ref": "hg19",
        "index": "ACTGGTGTCG-ACAGGACTTG",
        "description": "",
        "sample_name": "814206",
        "control": "N",
        "recipe": "R1",
        "operator": "script",
        "project": "814206",
    }
    return LimsFlowcellSample(**sample)


@pytest.fixture(name="lims_novaseq_samples_file")
def fixture_lims_novaseq_samples_file(raw_samples_dir: Path) -> Path:
    """Return the path to a file with sample info in lims format"""
    return raw_samples_dir / "raw_samplesheet_novaseq.json"


@pytest.fixture(name="lims_novaseq_samples")
def fixture_lims_novaseq_samples(lims_novaseq_samples_file: Path) -> List[LimsFlowcellSample]:
    """Return a list of parsed flowcell samples"""
    with open(lims_novaseq_samples_file, "r") as in_file:
        raw_samples: List[dict] = json.load(in_file)
        return [LimsFlowcellSample(**sample) for sample in raw_samples]


@pytest.fixture(name="novaseq_run_parameters_object")
def fixture_novaseq_run_parameters_object(novaseq_run_parameters: Path) -> RunParameters:
    return RunParameters(novaseq_run_parameters)


@pytest.fixture(name="novaseq_sample_sheet_object")
def fixture_novaseq_sample_sheet_object(
    flowcell_name: str,
    lims_novaseq_samples: List[LimsFlowcellSample],
    novaseq_run_parameters_object: RunParameters,
) -> SampleSheetCreator:
    return SampleSheetCreator(
        flowcell_id=flowcell_name,
        lims_samples=lims_novaseq_samples,
        run_parameters=novaseq_run_parameters_object,
    )
